from .adnl import AdnlTransport, AdnlTransportError, AdnlChannel, Node
from .dht import DhtNode, DhtError, DhtClient, DhtValueNotFoundError
from .overlay import OverlayTransport, OverlayNode, OverlayTransportError



import typing

from ..liteclient.client import LiteClient
from pytoniq_core.boc.cell import Cell
from pytoniq_core.boc.address import Address
from pytoniq_core.tlb.account import StateInit, Account, SimpleAccount, ShardAccount
from pytoniq_core.tlb.block import CurrencyCollection
from pytoniq_core.tlb.transaction import ExternalMsgInfo, MessageAny, InternalMsgInfo


class ContractError(BaseException):
    pass


class Contract:

    def __init__(self, provider: LiteClient, address: Address, account: typing.Optional[Account] = None,
                 shard_account: typing.Optional[ShardAccount] = None, state_init: typing.Optional[StateInit] = None,
                 **kwargs):
        """
        :param provider: LiteClient
        :param address: contract address
        :param account: full account state
        :param shard_account: shard account
        :param state_init: account state init. usually used for contract deploying
        :param kwargs: some additional account attributes. for e.g. private key for wallet contracts
        """
        self.provider = provider
        self.address = address
        self.raw_account: typing.Optional[Account] = None
        self.account: SimpleAccount = None
        self.shard_account: typing.Optional[ShardAccount] = None
        self.is_active: bool = None
        self.is_uninitialized: bool = None
        self.is_frozen: bool = None
        self.state: typing.Optional[StateInit] = None
        self.state_init: typing.Optional[StateInit] = state_init
        self.set_account_attributes(account, shard_account)

        for k, v in kwargs.items():
            setattr(self, k, v)

    def set_account_attributes(self, account: typing.Optional[Account], shard_account: typing.Optional[ShardAccount]):
        self.raw_account = account
        self.account = SimpleAccount.from_raw(account, self.address)
        self.shard_account = shard_account
        self.is_active = self.account.is_active()
        self.is_uninitialized = self.account.is_uninitialized()
        self.is_frozen = self.account.is_frozen()

        self.state = None
        if self.is_active:
            self.state = self.account.state.state_init
        else:
            self.state = self.state_init

    @property
    def data(self) -> Cell:
        return self.state.data

    @property
    def code(self) -> Cell:
        return self.state.code

    @property
    def balance(self) -> int:
        """
        :return: balance from saved account state
        """
        return self.account.balance

    @classmethod
    async def from_address(cls, provider: LiteClient, address: typing.Union[str, Address], **kwargs):
        if isinstance(address, str):
            address = Address(address)
        account, shard_account = await provider.raw_get_account_state(address)
        return cls(provider=provider, address=address, account=account, shard_account=shard_account, **kwargs)

    @classmethod
    async def from_state_init(cls, provider: LiteClient, workchain: int, state_init: StateInit, **kwargs):
        address = Address((workchain, state_init.serialize().hash))
        return await cls.from_address(provider=provider, address=address, state_init=state_init, **kwargs)

    @classmethod
    async def from_code_and_data(cls, provider: LiteClient, workchain: int, code: Cell, data: Cell, **kwargs):
        state_init = StateInit(code=code, data=data)
        return await cls.from_state_init(provider=provider, workchain=workchain, state_init=state_init, **kwargs)

    async def update(self):
        account, shard_account = await self.raw_get_account_state()
        self.set_account_attributes(account, shard_account)

    async def raw_get_account_state(self):
        return (await self.provider.raw_get_account_state(address=self.address))[0]

    async def get_account_state(self):
        return await self.provider.get_account_state(self.address)

    async def get_balance(self) -> int:
        """
        :return: balance from current account state. better to
            await contract.update()
            contract.balance
        """
        account_state = await self.get_account_state()
        return account_state.balance

    async def run_get_method(self, method: typing.Union[str, int], stack: typing.Optional[list] = None):
        if stack is None:
            stack = []
        return await self.provider.run_get_method(self.address, method, stack)

    @staticmethod
    def create_external_msg(src: typing.Optional[Address] = None, dest: typing.Optional[Address] = None,
                            import_fee: int = 0, state_init: typing.Optional[StateInit] = None,
                            body: Cell = None) -> MessageAny:
        info = ExternalMsgInfo(src, dest, import_fee)
        if body is None:
            body = Cell.empty()
        message = MessageAny(info=info, init=state_init, body=body)
        return message

    @staticmethod
    def create_internal_msg(ihr_disabled: bool = True, bounce: bool = None, bounced: bool = False, src: Address = None,
                            dest: Address = None,
                            value: typing.Union[CurrencyCollection, int] = 0, ihr_fee: int = 0, fwd_fee: int = 0,
                            created_lt: int = 0,
                            created_at: int = 0, state_init: typing.Optional[StateInit] = None,
                            body: Cell = None) -> MessageAny:
        if isinstance(value, int):
            value = CurrencyCollection(grams=value, other=None)
        if bounce is None:
            bounce = dest.is_bounceable
        info = InternalMsgInfo(ihr_disabled, bounce, bounced, src, dest, value, ihr_fee, fwd_fee, created_lt, created_at)
        if body is None:
            body = Cell.empty()
        message = MessageAny(info=info, init=state_init, body=body)
        return message

    async def send_external(self, src: typing.Optional[Address] = None, import_fee: int = 0,
                            state_init: typing.Optional[StateInit] = None, body: Cell = None):
        message = self.create_external_msg(src=src, dest=self.address, import_fee=import_fee, state_init=state_init,
                                           body=body)
        return await self.provider.raw_send_message(message.serialize().to_boc())

    async def send_init_external(self):
        if not self.state_init:
            raise ContractError('contract does not have state_init attribute')
        return await self.send_external(state_init=self.state_init)

    async def deploy_via_external(self):
        return await self.send_init_external()

    def __repr__(self):
        return f'<{self.account.state.type_} {self.__class__.__name__} {self.address}>'  # <active WalletV4R2 EQBvW8Z5huBkMJYdnfAEM5JqTNkuWX3diqYENkWsIL0XggGG>



import logging
import os
import time
import typing

from pytoniq_core.tl.block import BlockIdExt


logger = logging.getLogger('sync')


async def sync(client, to_block: BlockIdExt, init_block: BlockIdExt):
    logger.info(msg=f'syncing to {to_block}')
    from .client import LiteClient
    client: LiteClient
    valid_key_block_stored = False
    blocks_data = get_last_stored_blocks(init_block.root_hash.hex())
    if not blocks_data:
        logger.debug(f'no last blocks were found, syncing from the init block {init_block}')
        mc_block = init_block
        key_block = init_block
    else:
        ttl, key_ts, key_block, mc_block = parse_blocks(blocks_data)
        logger.debug(f'found key block with ttl {ttl}')
        valid_key_block_stored = True
        if ttl <= time.time():
            logger.debug(f'key block ttl has been expired, syncing from the init block {init_block}')
            mc_block = init_block
            key_block = init_block
            valid_key_block_stored = False

    """
    Looks like last mc block should be be much sooner than last synced key block
    (cause we store key blocks with big ttl), but the lite node doesnt have to store any blocks except key blocks which
    persistent state has not expired.
    So the best solution is to ask liteserver if it remembers stored mc block and if not sync from the key block (or even init). 
    """
    try:
        best_key, best_key_ts = await client.get_mc_block_proof(known_block=mc_block, target_block=to_block, return_best_key_block=True)
    except:  # TODO specify exception class
        best_key, best_key_ts = await client.get_mc_block_proof(known_block=key_block, target_block=to_block, return_best_key_block=True)

    if valid_key_block_stored:
        best_key, best_key_ts = choose_key_block(key_block, key_ts, best_key, best_key_ts)

    key_ttl = persistent_state_ttl(best_key_ts)
    logger.info(msg=f'synced! store key block {best_key} with ttl {key_ttl}')

    store_blocks(blocks_to_bytes(key_ttl, best_key_ts, best_key, to_block), True, init_block.root_hash.hex())
    return True


def get_block_store_path():
    dir = os.path.join(os.path.curdir, '.blockstore')
    path = os.path.normpath(dir)
    if not os.path.isdir(path):
        os.mkdir(path)
    return path


def get_last_stored_blocks(init_block_hash: str) -> typing.Optional[bytes]:
    path = get_block_store_path()
    files = []
    for f in os.listdir(path):
        if init_block_hash in f:
            files.append(f)
    if not len(files):
        return None
    with open(os.path.join(path, files[-1]), 'rb') as f:
        result = f.read()
    return result


def store_blocks(data: bytes, delete_old: bool = True, init_block_hash: str = None):
    path = get_block_store_path()
    if delete_old:
        for f in os.listdir(path):
            deleted = False
            if init_block_hash:
                if init_block_hash in f:
                    os.remove(os.path.join(path, f))
                    deleted = True
            if not deleted:
                ttl = int(f[:8], 16)
                if ttl < time.time():
                    os.remove(os.path.join(path, f))
    file_name = data[:88].hex() + init_block_hash + '.blks'
    with open(os.path.join(path, file_name), 'wb') as f:
        f.write(data)
    return


def parse_blocks(data: bytes) -> typing.Tuple[int, int, BlockIdExt, BlockIdExt]:
    ttl = int.from_bytes(data[:4], 'big', signed=False)
    ts = int.from_bytes(data[4:8], 'big', signed=False)
    last_trusted_key_block = BlockIdExt.from_bytes(data[8:88])
    last_trusted_mc_block = BlockIdExt.from_bytes(data[88:])
    return ttl, ts, last_trusted_key_block, last_trusted_mc_block


def blocks_to_bytes(ttl: int, ts: int, last_trusted_key_block: BlockIdExt, last_trusted_mc_block: BlockIdExt):
    return ttl.to_bytes(4, 'big', signed=False) + ts.to_bytes(4, 'big', signed=False) + last_trusted_key_block.to_bytes() + last_trusted_mc_block.to_bytes()


def count_trailing_zeros(x: int):
    return (x & -x).bit_length() - 1


def persistent_state_ttl(ts: int):
    # https://github.com/ton-blockchain/ton/blob/d2b418bb703ed6ccd89b7d40f9f1e44686012014/validator/interfaces/validator-manager.h#L176
    x = ts / (1 << 17)
    assert x > 0
    b = count_trailing_zeros(int(x))
    return ts + ((1 << 18) << b)


def choose_key_block(blk: BlockIdExt, blk_ts: int, other_blk: typing.Optional[BlockIdExt], other_ts: typing.Optional[int]):
    if other_blk is None:
        return blk, blk_ts
    if blk is None:
        return other_blk, other_ts
    p1 = persistent_state_ttl(blk_ts)
    p2 = persistent_state_ttl(other_ts)
    c_t = time.time()
    if p1 < c_t and p2 < c_t:
        if blk.seqno > other_blk.seqno:
            return blk, blk_ts
        else:
            return other_blk, other_ts
    if p1 < c_t:
        return other_blk, other_ts
    if p2 < c_t:
        return blk, blk_ts

    d1 = p1 - c_t
    d2 = p2 - c_t

    min_time = 21 * 3600 * 24  # 3 weeks
    if d1 >= min_time:
        if d2 < min_time:
            return blk, blk_ts
        else:
            if blk.seqno > other_blk.seqno:
                return blk, blk_ts
            else:
                return other_blk, other_ts
    elif d2 >= min_time:
        return other_blk, other_ts
    else:
        if blk.seqno > other_blk.seqno:
            return blk, blk_ts
        return other_blk, other_ts



import asyncio
import base64
import inspect
import logging
import random
import time
import hashlib
import typing
from asyncio import transports
from typing import Any

from pytoniq_core.tl.generator import TlGenerator

from pytoniq_core.crypto.ciphers import Server, Client, AdnlChannel, get_random, aes_ctr_encrypt, aes_ctr_decrypt, get_shared_key, create_aes_ctr_sipher_from_key_n_data


class SocketProtocol(asyncio.DatagramProtocol):

    def __init__(self, timeout: int = 10):
        # https://github.com/eerimoq/asyncudp/blob/main/asyncudp/__init__.py
        self._error = None
        self._packets = asyncio.Queue(10000)
        self.timeout = timeout
        self.logger = logging.getLogger(self.__class__.__name__)

    def connection_made(self, transport: transports.DatagramTransport) -> None:
        super().connection_made(transport)

    def datagram_received(self, data: bytes, addr: typing.Tuple[typing.Union[str, Any], int]) -> None:
        self.logger.debug(f'received {len(data)} bytes')
        self._packets.put_nowait((data, addr))
        super().datagram_received(data, addr)

    def error_received(self, exc: Exception) -> None:
        raise exc
        super().error_received(exc)

    async def receive(self):
        return await self._packets.get()


class Node(Server):

    def __init__(
            self,
            peer_host: str,  # ipv4 host
            peer_port: int,  # port
            peer_pub_key: str,
            transport: "AdnlTransport"
    ):
        self.host = peer_host
        self.port = peer_port
        super().__init__(peer_host, peer_port, base64.b64decode(peer_pub_key))
        self.channels: typing.List[AdnlChannel] = []
        self.seqno = 1
        self.confirm_seqno = 0
        self.key_id = self.get_key_id()
        self.transport = transport
        self.pinger: asyncio.Task = None
        self.connected = False
        self.logger = logging.getLogger(self.__class__.__name__)
        self._lost_pings = 0

    async def connect(self):
        return await self.transport.connect_to_peer(self)

    async def send_ping(self) -> None:
        random_id = get_random(8)
        resp = await self.transport.send_query_message(tl_schema_name='dht.ping', data={'random_id': random_id}, peer=self)
        assert resp[0].get('random_id') == int.from_bytes(random_id, 'big', signed=True)

    def start_ping(self):
        self.pinger = asyncio.create_task(self.ping())

    async def ping(self):
        while True:
            try:
                await self.send_ping()
            except asyncio.TimeoutError:
                self._lost_pings += 1
                if self._lost_pings > 3:
                    self.transport.peers.pop(self.key_id)
                    await self.disconnect()
                else:
                    continue
            self.logger.debug(f'pinged {self.key_id.hex()}')
            await asyncio.sleep(3)

    async def get_signed_address_list(self):
        return (await self.transport.send_query_message('dht.getSignedAddressList', {}, self))[0]

    @property
    def addr(self) -> typing.Tuple[str, int]:
        """
        :return: ipv4 node address as (host, port)
        """
        return self.host, self.port

    def inc_seqno(self):
        self.seqno += 1

    async def disconnect(self):
        if self.connected:
            self.connected = False
            self.pinger.cancel()


class AdnlTransportError(Exception):
    pass


class AdnlTransport:

    def __init__(self,
                 private_key: bytes = None,
                 tl_schemas_path: str = None,
                 local_address: tuple = ('0.0.0.0', None),
                 *args, **kwargs
                 ) -> None:
        """
        ADNL Transport abstract class
        """

        """########### init ###########"""
        self.loop: asyncio.AbstractEventLoop = None
        self.timeout = kwargs.get('timeout', 10)
        self.listener: asyncio.Task = None
        self.logger = logging.getLogger(self.__class__.__name__)
        self.tasks: typing.Dict[str, asyncio.Future] = {}
        self.query_handlers: typing.Dict[str, typing.Callable] = {}
        self.custom_handlers: typing.Dict[str, typing.Callable] = {}
        self._message_parts: typing.Dict[str, dict] = {}  # {'hash': {'remained': int, 'parts': list}}

        """########### connection ###########"""
        self.transport: asyncio.DatagramTransport = None
        self.protocol: SocketProtocol = None
        if local_address[1] is None:
            local_address = local_address[0], random.randint(10000, 60000)
        self.local_address = local_address

        """########### TL ###########"""
        if tl_schemas_path is None:
            self.schemas = TlGenerator.with_default_schemas().generate()
        else:
            self.schemas = TlGenerator(tl_schemas_path).generate()
        self.adnl_query_sch = self.schemas.get_by_name('adnl.message.query')
        self.adnl_packet_content_sch = self.schemas.get_by_name('adnl.packetContents')
        self.create_channel_sch = self.schemas.get_by_name('adnl.message.createChannel')

        """########### crypto ###########"""
        self.peers: typing.Dict[bytes, Node] = {}
        if private_key is None:
            private_key = Client.generate_ed25519_private_key()
        self.client = Client(private_key)
        self.local_id = self.client.get_key_id()
        self.channels: typing.Dict[bytes, AdnlChannel] = {}
        self.enc_sipher = None
        self.dec_sipher = None

    @staticmethod
    def _get_rand():
        rand = get_random(16)
        if rand[0] & 1 > 0:
            return rand[1:]
        return rand[1:8]

    @staticmethod
    def compute_flags_for_packet(data: dict) -> dict:
        """
        :param data: dict with TL Scheme arguments
        :return: data with computed flags field
        """
        flags = 0
        if 'from' in data:
            flags += 1 << 0
        if 'from_short' in data:
            flags += 1 << 1
        if 'message' in data:
            flags += 1 << 2
        if 'messages' in data:
            flags += 1 << 3
        if 'address' in data:
            flags += 1 << 4
        if 'priority_address' in data:
            flags += 1 << 5
        if 'seqno' in data:
            flags += 1 << 6
        if 'confirm_seqno' in data:
            flags += 1 << 7
        if 'recv_addr_list_version' in data:
            flags += 1 << 8
        if 'recv_priority_addr_list_version' in data:
            flags += 1 << 9
        if 'reinit_date' in data or 'dst_reinit_date' in data:
            flags += 1 << 10
        if 'signature' in data:
            flags += 1 << 11

        return data | {'flags': flags}

    def _prepare_packet_content_msg(self, data: dict, peer: Node = None) -> dict:
        """
        Adds random bytes, seqno, confirm_seqno and flags in message args if they were not provided
        """
        if 'rand1' not in data or 'rand2' not in data:
            data['rand1'] = self._get_rand()
            data['rand2'] = self._get_rand()

        if data.get('seqno') is None:
            if peer is None:
                raise AdnlTransportError('Must either specify seqno in data or provide peer to method')
            data['seqno'] = peer.seqno
        if data.get('confirm_seqno') is None:
            if peer is None:
                raise AdnlTransportError('Must either specify confirm_seqno in data or provide peer to method')
            data['confirm_seqno'] = peer.confirm_seqno

        return self.compute_flags_for_packet(data)

    def _decrypt_any(self, resp_packet: bytes) -> typing.Tuple[bytes, typing.Optional[Node]]:
        """
        :param resp_packet: bytes of received packet
        :return: decrypted packet and maybe `Node`
        """
        key_id = resp_packet[:32]
        if key_id == self.client.get_key_id():
            server_public_key = resp_packet[32:64]
            checksum = resp_packet[64:96]
            encrypted = resp_packet[96:]

            peer_crypto = Server('', 0, server_public_key)

            shared_key = get_shared_key(self.client.x25519_private.encode(),
                                        peer_crypto.x25519_public.encode())

            dec_cipher = create_aes_ctr_sipher_from_key_n_data(shared_key, checksum)
            decrypted = aes_ctr_decrypt(dec_cipher, encrypted)
            assert hashlib.sha256(decrypted).digest() == checksum, 'invalid checksum'
            return decrypted, None
        else:
            for peer_id, channel in self.channels.items():
                if key_id == channel.server_aes_key_id:
                    checksum = resp_packet[32:64]
                    encrypted = resp_packet[64:]
                    decrypted = channel.decrypt(encrypted, checksum)
                    assert hashlib.sha256(decrypted).digest() == checksum, 'invalid checksum'
                    return decrypted, self.peers.get(peer_id)
            # TODO make new connection
            self.logger.debug(f'unknown key id from node: {key_id.hex()}')
            return b'', None

    def _process_outcoming_message(self, message: dict) -> typing.Optional[asyncio.Future]:
        future = self.loop.create_future()
        type_ = message['@type']
        if type_ == 'adnl.message.query':
            self.tasks[message.get('query_id')[::-1].hex()] = future
        elif type_ == 'adnl.message.createChannel':
            self.tasks[message.get('key')] = future
        else:
            return
        return future

    def _create_futures(self, data: dict) -> typing.List[asyncio.Future]:
        futures = []
        if data.get('message'):
            future = self._process_outcoming_message(data['message'])
            if future is not None:
                futures.append(future)

        if data.get('messages'):
            for message in data['messages']:
                future = self._process_outcoming_message(message)
                if future is not None:
                    futures.append(future)
        return futures

    @staticmethod
    async def _receive(futures: typing.List[asyncio.Future]) -> list:
        return list(await asyncio.gather(*futures))

    async def _process_incoming_message(self, message: dict, peer: Node):
        if peer:
            self.logger.debug(f'Received message {message} from peer {peer.get_key_id().hex()}')
        if message['@type'] == 'adnl.message.answer':
            future = self.tasks.pop(message.get('query_id'))
            future.set_result(message['answer'])
        elif message['@type'] == 'adnl.message.confirmChannel':
            if message.get('peer_key') in self.tasks:
                future = self.tasks.pop(message.get('peer_key'))
                future.set_result(message)
        elif message['@type'] == 'adnl.message.query':
            if peer is None:
                self.logger.info(f'Received query message from unknown peer: {message}')
                # not implemented, todo: make connection with new peer
                return
            await self._process_query_message(message, peer)
        elif message['@type'] == 'adnl.message.custom':
            if peer is None:
                # should not ever happen fixme
                self.logger.info(f'Received custom message from unknown peer: {message}')
                return
            await self._process_custom_message(message, peer)
        elif message['@type'] == 'adnl.message.part':
            hash_ = message['hash']
            if hash_ not in self._message_parts:
                self._message_parts[hash_] = {'remained': message['total_size'], 'parts': []}

            self._message_parts[hash_]['remained'] -= len(message['data'])
            self._message_parts[hash_]['parts'].append(message)

            if self._message_parts[hash_]['remained'] == 0:
                data = self._collect_adnl_message_parts(hash_)
                if isinstance(data, dict) and data['@type'] != 'adnl.message.part':  # to avoid infinity recursion, but should never happen
                    await self._process_incoming_message(data, peer)
        else:
            self.logger.info(f'unexpected message type received: {message}')
            # raise AdnlTransportError(f'unexpected message type received: {message}')

    def _collect_adnl_message_parts(self, hash_: str, deserialize_after: bool = True):
        if hash_ not in self._message_parts:
            raise AdnlTransportError(f'Provided hash not in message parts')
        parts = sorted(self._message_parts[hash_]['parts'], key=lambda i: i['offset'])
        full_data = b''
        for part in parts:
            full_data += part['data']
        if deserialize_after:
            return self.schemas.deserialize(full_data)[0]

    async def _process_query_message(self, message: dict, peer: Node):
        query = message.get('query')
        # it's divided into separate method because some higher level protocols over ADNL need specific query processing
        await self._process_query_handler(message, query, peer)

    async def _process_query_handler(self, message: dict, query: dict, peer: Node):
        # try to get handler for specific query and if there is no, try to get default handler
        handler = self.query_handlers.get(query['@type'], self.query_handlers.get(None))
        if handler:
            if inspect.iscoroutinefunction(handler):
                response = await handler(query)
            else:
                response = handler(query)
            if response is not None:
                await self.send_answer_message(response, message.get('query_id'), peer)

    async def _process_custom_message(self, message: dict, peer: Node):
        data = message.get('data')
        await self._process_custom_message_handler(data, peer)

    async def _process_custom_message_handler(self, data: dict, peer: Node):
        handler = self.custom_handlers.get(data['@type'], self.custom_handlers.get(None))
        if handler:
            if inspect.iscoroutinefunction(handler):
                response = await handler(data)
            else:
                response = handler(data)
            if response is not None:
                await self.send_custom_message(response, peer)

    def set_query_handler(self, type_: str, handler: typing.Callable) -> None:
        """
        :param type_: TL type of message
        :param handler: function to handle message. **Must** return dict or bytes or None. If
            None returned than answer won't be sent to the sender
        :return:
        """
        self.query_handlers[type_] = handler

    def set_default_query_handler(self, handler: typing.Callable):
        """
        Same as `set_query_handler` when there is no handlers for query specific type.
        :param handler:
        :return:
        """
        self.set_query_handler(None, handler)

    def set_custom_message_handler(self, type_: str, handler: typing.Callable):
        """
        :param type_: TL type of message
        :param handler: function to handle message. **Must** return dict or bytes or None. If
            None returned than answer won't be sent to the sender.
        :return:
        """
        self.custom_handlers[type_] = handler

    def set_default_custom_message_handler(self, handler: typing.Callable):
        """
        Same as `set_custom_message_handler` when there is no handlers for query specific type.
        :param handler:
        :return:
        """
        self.set_custom_message_handler(None, handler)

    async def listen(self):
        while True:
            packet, addr = await self.protocol.receive()

            decrypted, peer = self._decrypt_any(packet)
            if not decrypted:
                continue
            response = self.schemas.deserialize(decrypted)[0]

            if peer is None:
                if 'from_short' in response:
                    peer = self.peers.get(bytes.fromhex(response['from_short']['id']))

            if peer is not None:
                received_confirm_seqno = response.get('confirm_seqno', 0)
                if received_confirm_seqno > peer.confirm_seqno:
                    peer.confirm_seqno = received_confirm_seqno

            message = response.get('message')
            messages = response.get('messages')

            if message:
                await self._process_incoming_message(message, peer)
            if messages:
                for message in messages:
                    await self._process_incoming_message(message, peer)

    async def send_message_in_channel(self, data: dict, channel: typing.Optional[AdnlChannel] = None, peer: Node = None) -> list:

        if peer is None:
            raise AdnlTransportError('Must provide peer')

        data = self._prepare_packet_content_msg(data, peer)
        sending_seqno = data.get('seqno')

        futures = self._create_futures(data)

        if channel is None:
            if not len(peer.channels):
                raise AdnlTransportError(f'Peer has no channels and channel was not provided')
            channel = peer.channels[0]

        if peer.seqno == sending_seqno:
            peer.inc_seqno()
        else:
            raise Exception(f'sending seqno {sending_seqno}, client seqno: {peer.seqno}')
        serialized = self.schemas.serialize(self.adnl_packet_content_sch, data)
        res = channel.encrypt(serialized)

        self.transport.sendto(res, addr=peer.addr)
        result = await asyncio.wait_for(self._receive(futures), self.timeout)

        return result

    async def send_message_outside_channel(self, data: dict, peer: Node) -> list:
        """
        Serializes, signs and encrypts sending message.
        :param peer: peer
        :param data: data for `adnl.packetContents` TL Scheme
        :return: decrypted and deserialized response
        """
        data = self._prepare_packet_content_msg(data, peer)
        sending_seqno = data.get('seqno')

        data = self.compute_flags_for_packet(data)

        futures = self._create_futures(data)

        serialized1 = self.schemas.serialize(self.adnl_packet_content_sch, self.compute_flags_for_packet(data))
        signature = self.client.sign(serialized1)
        serialized2 = self.schemas.serialize(self.adnl_packet_content_sch,
                                             self.compute_flags_for_packet(data | {'signature': signature}))

        checksum = hashlib.sha256(serialized2).digest()
        shared_key = get_shared_key(self.client.x25519_private.encode(), peer.x25519_public.encode())
        init_cipher = create_aes_ctr_sipher_from_key_n_data(shared_key, checksum)
        data = aes_ctr_encrypt(init_cipher, serialized2)

        res = peer.get_key_id() + self.client.ed25519_public.encode() + checksum + data
        self.transport.sendto(res, addr=(peer.host, peer.port))

        if peer.seqno == sending_seqno:
            peer.inc_seqno()
        else:
            raise Exception(f'sending seqno {sending_seqno}, client seqno: {peer.seqno}')
        if futures:
            result = await asyncio.wait_for(self._receive(futures), self.timeout)
            return result

    async def start(self):
        self.loop = asyncio.get_running_loop()
        self.transport, self.protocol = await self.loop.create_datagram_endpoint(
            lambda: SocketProtocol(timeout=self.timeout),
            local_addr=self.local_address,
            reuse_port=True
        )
        self.listener = self.loop.create_task(self.listen())
        return

    def _get_default_message(self):
        return {
            '@type': 'adnl.message.query',
            'query_id': get_random(32),
            'query': self.schemas.get_by_name('dht.getSignedAddressList').little_id()
        }

    async def connect_to_peer(self, peer: Node) -> list:
        """
        Connects to the peer, creates channel and asks for a signed list in channel.
        :param peer: peer connect to
        :return: response dict for default message
        """

        ts = int(time.time())
        channel_client = Client(Client.generate_ed25519_private_key())
        create_channel_message = {
            '@type': 'adnl.message.createChannel',
            'key': channel_client.ed25519_public.encode().hex(),
            'date': ts
        }

        default_message = self._get_default_message()

        from_ = self.schemas.serialize(self.schemas.get_by_name('pub.ed25519'), data={'key': self.client.ed25519_public.encode().hex()})
        data = {
            'from': from_,
            # 'from_short': {'id': self.client.get_key_id().hex()},
            'messages': [create_channel_message, default_message],
            'address': {
                'addrs': [],
                'version': ts,
                'reinit_date': ts,
                'priority': 0,
                'expire_at': 0,
            },
            'recv_addr_list_version': ts,
            'reinit_date': ts,
            'dst_reinit_date': 0,
        }

        messages = await self.send_message_outside_channel(data, peer)
        confirm_channel = messages[0]
        assert confirm_channel.get('@type') == 'adnl.message.confirmChannel', (f'expected adnl.message.confirmChannel,'
                                                                               f' got {confirm_channel.get("@type")}')
        assert confirm_channel['peer_key'] == channel_client.ed25519_public.encode().hex()

        channel_peer = Server(peer.host, peer.port, bytes.fromhex(confirm_channel['key']))
        channel = AdnlChannel(channel_client, channel_peer, self.local_id, peer.get_key_id())
        self.channels[peer.get_key_id()] = channel
        peer.channels.append(channel)

        peer.start_ping()
        peer.connected = True
        self.peers[peer.key_id] = peer

        return messages[1]

    async def close(self):
        self.listener.cancel()
        while not self.listener.cancelled():
            await asyncio.sleep(0)
        self.transport.abort()

    async def send_query_message(self, tl_schema_name: str, data: dict, peer: Node) -> typing.List[dict]:
        message = {
            '@type': 'adnl.message.query',
            'query_id': get_random(32),
            'query': self.schemas.serialize(
                    self.schemas.get_by_name(tl_schema_name),
                    data
            )
        }

        data = {
            'message': message,
        }

        result = await self.send_message_in_channel(data, None, peer)
        return result

    async def send_answer_message(self, response: typing.Union[dict, bytes], query_id: bytes, peer: Node):
        message = {
            '@type': 'adnl.message.answer',
            'query_id': query_id,
            'answer': response
        }

        data = {
            'message': message,
        }
        return await self.send_message_in_channel(data, None, peer)

    async def send_custom_message(self, message: typing.Union[dict, bytes], peer: Node) -> list:

        custom_message = {
            '@type': 'adnl.message.custom',
            'data': message
        }

        data = {
            'message': custom_message,
        }

        result = await self.send_message_in_channel(data, None, peer)
        return result



import time
import hashlib
import typing

from pytoniq_core.tl.generator import TlGenerator

from pytoniq_core.crypto.ciphers import get_random

from .adnl import Node, AdnlTransport, AdnlTransportError


class OverlayTransportError(AdnlTransportError):
    pass


class OverlayNode(Node):

    def __init__(
            self,
            peer_host: str,  # ipv4 host
            peer_port: int,  # port
            peer_pub_key: str,
            transport: "OverlayTransport"
    ):
        self.transport: "OverlayTransport" = None
        super().__init__(peer_host, peer_port, peer_pub_key, transport)

    async def send_ping(self) -> None:
        peers = [
            self.transport.get_signed_myself()
        ]
        await self.transport.send_query_message('overlay.getRandomPeers', {'peers': {'nodes': peers}}, peer=self)


class OverlayTransport(AdnlTransport):

    def __init__(self,
                 private_key: bytes = None,
                 tl_schemas_path: str = None,
                 local_address: tuple = ('0.0.0.0', None),
                 overlay_id: typing.Union[str, bytes] = None,
                 *args, **kwargs
                 ) -> None:

        super().__init__(private_key, tl_schemas_path, local_address, *args, **kwargs)
        if overlay_id is None:
            raise OverlayTransportError('must provide overlay id in OverlayTransport')

        if isinstance(overlay_id, bytes):
            overlay_id = overlay_id.hex()

        self.overlay_id = overlay_id

    @staticmethod
    def get_overlay_id(zero_state_file_hash: typing.Union[bytes, str],
                       workchain: int = 0, shard: int = -9223372036854775808) -> str:

        if isinstance(zero_state_file_hash, bytes):
            zero_state_file_hash = zero_state_file_hash.hex()

        schemes = TlGenerator.with_default_schemas().generate()

        sch = schemes.get_by_name('tonNode.shardPublicOverlayId')
        data = {
            "workchain": workchain,
            "shard": shard,
            "zero_state_file_hash": zero_state_file_hash
        }

        key_id = hashlib.sha256(schemes.serialize(sch, data)).digest()

        sch = schemes.get_by_name('pub.overlay')
        data = {
            'name': key_id
        }

        key_id = schemes.serialize(sch, data)

        return hashlib.sha256(key_id).digest().hex()

    @classmethod
    def get_mainnet_overlay_id(cls, workchain: int = 0, shard: int = -9223372036854775808) -> str:
        return cls.get_overlay_id('5e994fcf4d425c0a6ce6a792594b7173205f740a39cd56f537defd28b48a0f6e', workchain, shard)

    @classmethod
    def get_testnet_overlay_id(cls, workchain: int = 0, shard: int = -9223372036854775808) -> str:
        return cls.get_overlay_id('67e20ac184b9e039a62667acc3f9c00f90f359a76738233379efa47604980ce8', workchain, shard)

    async def _process_query_message(self, message: dict, peer: OverlayNode):
        query = message.get('query')
        if isinstance(query, list):
            if query[0]['@type'] == 'overlay.query':
                assert query[0]['overlay'] == self.overlay_id, 'Unknown overlay id received'
            query = query[-1]
        await self._process_query_handler(message, query, peer)

    async def _process_custom_message(self, message: dict, peer: Node):
        data = message.get('data')
        if isinstance(data, list):
            if data[0]['@type'] == 'overlay.query':
                assert data[0]['overlay'] == self.overlay_id, 'Unknown overlay id received'
            data = data[-1]
        await self._process_custom_message_handler(data, peer)

    def get_signed_myself(self):
        ts = int(time.time())

        overlay_node_data = {'id': {'@type': 'pub.ed25519', 'key': self.client.ed25519_public.encode().hex()},
                             'overlay': self.overlay_id, 'version': ts, 'signature': b''}

        overlay_node_to_sign = self.schemas.serialize(self.schemas.get_by_name('overlay.node.toSign'),
                                                      {'id': {'id': self.client.get_key_id().hex()},
                                                       'overlay': self.overlay_id,
                                                       'version': overlay_node_data['version']})
        signature = self.client.sign(overlay_node_to_sign)

        overlay_node = overlay_node_data | {'signature': signature}
        return overlay_node

    async def send_query_message(self, tl_schema_name: str, data: dict, peer: Node) -> typing.List[dict]:

        message = {
            '@type': 'adnl.message.query',
            'query_id': get_random(32),
            'query': self.schemas.serialize(self.schemas.get_by_name('overlay.query'), data={'overlay': self.overlay_id})
                     + self.schemas.serialize(self.schemas.get_by_name(tl_schema_name), data)
        }
        data = {
            'message': message,
        }

        result = await self.send_message_in_channel(data, None, peer)
        return result

    def get_message_with_overlay_prefix(self, schema_name: str, data: dict) -> bytes:
        return (self.schemas.serialize(
                    schema=self.schemas.get_by_name('overlay.query'),
                    data={'overlay': self.overlay_id})
                + self.schemas.serialize(
                    schema=self.schemas.get_by_name(schema_name),
                    data=data)
                )

    def _get_default_message(self):
        peers = [
            self.get_signed_myself()
        ]
        return {
            '@type': 'adnl.message.query',
            'query_id': get_random(32),
            'query': self.get_message_with_overlay_prefix('overlay.getRandomPeers', {'peers': {'nodes': peers}})
        }

    async def get_random_peers(self, peer: OverlayNode):
        overlay_node = self.get_signed_myself()

        peers = [
            overlay_node
        ]
        return await self.send_query_message(tl_schema_name='overlay.getRandomPeers', data={'peers': {'nodes': peers}},
                                             peer=peer)

    async def get_capabilities(self, peer: OverlayNode):
        return await self.send_query_message(tl_schema_name='tonNode.getCapabilities', data={}, peer=peer)



from .contract import *
from .liteclient import *
from .adnl import *
from pytoniq_core.boc import *
from pytoniq_core.crypto import *
from pytoniq_core.proof import *
from pytoniq_core.tl import *
from pytoniq_core.tlb import *



import logging
import pytest

from pytoniq.adnl.adnl import AdnlTransport
from pytoniq.adnl.dht import DhtClient, DhtNode


adnl = AdnlTransport(timeout=5)


client = DhtClient.from_mainnet_config(adnl)


@pytest.mark.asyncio
async def test_dht():
    logging.basicConfig(level=logging.DEBUG)
    await adnl.start()

    foundation_adnl_addr = '516618cf6cbe9004f6883e742c9a2e3ca53ed02e3e36f4cef62a98ee1e449174'
    resp = await client.find_value(key=DhtClient.get_dht_key_id(bytes.fromhex(foundation_adnl_addr)))
    print(resp)

    assert resp['@type'] == 'dht.valueFound'



import base64
import hashlib
import logging
import asyncio
import random
import socket
import struct
import typing

import requests

from .sync import choose_key_block, sync
from .utils import init_mainnet_block, init_testnet_block
from pytoniq_core.boc import Slice, Cell
from pytoniq_core.proof.check_proof import check_block_header_proof, check_shard_proof, check_account_proof, check_proof, \
    check_block_signatures, compute_validator_set
from pytoniq_core.boc.address import Address

from pytoniq_core.crypto.ciphers import Server, Client, get_random, create_aes_ctr_cipher, aes_ctr_encrypt, aes_ctr_decrypt, get_shared_key
from pytoniq_core.crypto.crc import crc16

from pytoniq_core.tl.generator import TlGenerator, TlSchema
from pytoniq_core.tl.block import BlockIdExt, BlockId  # do not remove this import!
from pytoniq_core.tlb.config import ConfigParam34, ConfigParam28, ConfigParam
from pytoniq_core.tlb.transaction import Transaction
from pytoniq_core.tlb.utils import deserialize_shard_hashes

from pytoniq_core.tlb.vm_stack import VmStack
from pytoniq_core.tlb.block import Block, ShardDescr, BinTree, ShardStateUnsplit, KeyExtBlkRef
from pytoniq_core.tlb.account import Account, SimpleAccount, ShardAccount, AccountBlock


class LiteClientError(Exception):
    pass


class RunGetMethodError(LiteClientError):
    def __init__(self, address: typing.Any, method: typing.Any, exit_code: int):
        self.address = address
        self.method = method
        self.exit_code = exit_code
        super().__init__(f'get method "{method}" for account {address} returned exit code {exit_code}')


class LiteClient:

    def __init__(self,
                 host: str,  # ipv4 host
                 port: int,
                 server_pub_key: str,  # server ed25519 public key in base64,
                 timeout: int = 10,
                 tl_schemas_path: typing.Optional[str] = None,
                 trust_level: int = 1,
                 init_key_block: BlockIdExt = None,
                 ) -> None:
        """
        ADNL over TCP client for `liteservers` usage
        """

        """########### init ###########"""
        self.tasks = {}
        self.inited = False
        self.logger = logging.getLogger(self.__class__.__name__)
        self.timeout = timeout

        """########### sync ###########"""
        self.last_mc_block: BlockIdExt = None
        self.last_shard_blocks: typing.Dict[int, BlockIdExt] = None
        self.last_key_block: BlockIdExt = None
        self.trust_level = trust_level
        self.init_key_block: BlockIdExt = init_key_block
        if not self.trust_level and not init_key_block:
            raise LiteClientError('trust level is zero but no init block provided')

        """########### crypto ###########"""
        self.server = Server(host, port, base64.b64decode(server_pub_key))
        self.client = Client(Client.generate_ed25519_private_key())
        self.enc_sipher = None
        self.dec_sipher = None

        """########### connection ###########"""
        self.reader: asyncio.StreamReader = None
        self.writer: asyncio.StreamWriter = None
        self.loop: asyncio.AbstractEventLoop = None
        self.delta = 0.02  # listen delay

        self.listener: asyncio.Task = None
        rand_int = str(random.randint(10**9, 3 * 10**9))
        self.listener_name: str = 'listener' + rand_int
        self.pinger: asyncio.Task = None
        self.pinger_name: str = 'pinger' + rand_int
        self.updater: asyncio.Task = None
        self.updater_name: str = 'updater' + rand_int

        """########### TL ###########"""
        if tl_schemas_path is None:
            self.schemas = TlGenerator.with_default_schemas().generate()
        else:
            self.schemas = TlGenerator(tl_schemas_path).generate()
        # for better performance:
        self.ping_sch = self.schemas.get_by_name('tcp.ping')
        self.pong_sch = self.schemas.get_by_name('tcp.pong')
        self.adnl_query_sch = self.schemas.get_by_name('adnl.message.query')
        self.ls_query_sch = self.schemas.get_by_name('liteServer.query')

    def encrypt(self, data: bytes) -> bytes:
        return aes_ctr_encrypt(self.enc_sipher, data)

    def decrypt(self, data: bytes) -> bytes:
        return aes_ctr_decrypt(self.dec_sipher, data)

    async def send(self, data: bytes, qid: typing.Union[str, int, None]) -> asyncio.Future:
        future = self.loop.create_future()
        self.writer.write(data)
        await self.writer.drain()
        self.tasks[qid] = future
        return future

    async def send_and_encrypt(self, data: bytes, qid: str) -> asyncio.Future:
        future = self.loop.create_future()
        self.writer.write(self.encrypt(data))
        await self.writer.drain()
        self.tasks[qid] = future
        return future

    async def receive(self, data_len: int) -> bytes:
        data = await self.reader.readexactly(data_len)
        return data

    async def receive_and_decrypt(self, data_len: int) -> bytes:
        data = self.decrypt(await self.receive(data_len))
        return data

    async def listen(self) -> None:
        while True:
            while not self.tasks:
                await asyncio.sleep(self.delta)

            data_len_encrypted = await self.receive(4)
            data_len = int(self.decrypt(data_len_encrypted)[::-1].hex(), 16)

            self.logger.debug(msg=f'received {data_len // 8} bytes of data')

            data_encrypted = await self.receive(data_len)
            data_decrypted = self.decrypt(data_encrypted)
            # check hashsum
            assert hashlib.sha256(data_decrypted[:-32]).digest() == data_decrypted[-32:], 'incorrect checksum'
            result = self.deserialize_adnl_query(data_decrypted[:-32])

            if not result:
                # for handshake
                result = {}

            qid = result.get('query_id', result.get('random_id'))  # return query_id for ordinary requests, random_id for ping-pong requests, None for handshake

            request = self.tasks.pop(qid)

            result = result.get('answer', {})
            request.set_result(result)

    async def connect(self) -> None:
        self.loop = asyncio.get_running_loop()
        handshake = self.handshake()
        self.reader, self.writer = await asyncio.wait_for(asyncio.open_connection(self.server.host, self.server.port), self.timeout)
        future = await asyncio.wait_for(self.send(handshake, None), self.timeout)
        self.listener = asyncio.create_task(self.listen(), name=self.listener_name)
        await self.update_last_blocks()
        self.pinger = asyncio.create_task(self.ping(), name=self.pinger_name)
        self.updater = asyncio.create_task(self.block_updater(), name=self.updater_name)
        await future
        self.inited = True

    async def reconnect(self) -> None:
        await self.close()
        await self.connect()

    async def close(self) -> None:
        for i in asyncio.all_tasks(self.loop):
            if i.get_name() in {self.listener_name, self.pinger_name, self.updater_name}:
                i.cancel()
                while not i.cancelled():
                    await asyncio.sleep(0.1)
        self.inited = False
        self.tasks = {}
        self.reader = None
        self.writer.close()
        await self.writer.wait_closed()
        self.writer = None
        self.logger.info(msg='client has been closed')

    def handshake(self) -> bytes:
        rand = get_random(160)
        self.dec_sipher = create_aes_ctr_cipher(rand[0:32], rand[64:80])
        self.enc_sipher = create_aes_ctr_cipher(rand[32:64], rand[80:96])
        checksum = hashlib.sha256(rand).digest()
        shared_key = get_shared_key(self.client.x25519_private.encode(), self.server.x25519_public.encode())
        init_cipher = create_aes_ctr_cipher(shared_key[0:16] + checksum[16:32], checksum[0:4] + shared_key[20:32])
        data = aes_ctr_encrypt(init_cipher, rand)
        return self.server.get_key_id() + self.client.ed25519_public.encode() + checksum + data

    @staticmethod
    def serialize_packet(data: bytes):
        result = (len(data) + 64).to_bytes(4, 'little')
        result += get_random(32)  # nonce
        result += data  # useful data
        result += hashlib.sha256(result[4:]).digest()  # hashsum
        return result

    def serialize_adnl_ls_query(self, schema: TlSchema, data: dict) -> typing.Tuple[bytes, str]:
        """
        :param schema: TL schema
        :param data: dict
        :return: result_bytes, qid
        """
        qid = get_random(32)
        res = self.schemas.serialize(
            self.adnl_query_sch,
            {'query_id': qid,
             'query': self.schemas.serialize(self.ls_query_sch,
                                             {'data': self.schemas.serialize(schema, data)}
                                             )
             }
        )
        return res, qid[::-1].hex()

    def deserialize_adnl_query(self, data: bytes) -> dict:
        return self.schemas.deserialize(data[32:], boxed=True)[0]

    def get_ping_query(self):
        ping_sch = self.schemas.get_by_name('tcp.ping')
        query_id = get_random(8)
        data = self.schemas.serialize(ping_sch, {'random_id': query_id})
        data = self.serialize_packet(data)
        ping_result = self.encrypt(data)
        return ping_result, int.from_bytes(query_id, 'big', signed=True)

    async def ping(self):
        while True:
            await asyncio.sleep(3)
            ping_query, qid = self.get_ping_query()
            pong = await self.send(ping_query, qid)
            await pong
            self.logger.debug(msg=f'ping - pong')

    async def liteserver_query(self, query: bytes, qid: str) -> dict:
        data = self.serialize_packet(query)
        resp = await self.send_and_encrypt(data, qid)
        await asyncio.wait_for(resp, self.timeout)
        result = resp.result()

        if 'code' in result and 'message' in result:
            raise LiteClientError(f'LiteClient crashed with {result["code"]} code. Message: {result["message"]}')

        return resp.result()

    async def liteserver_request(self, tl_schema_name: str, data: dict) -> dict:
        schema = self.schemas.get_by_name('liteServer.' + tl_schema_name)
        self.logger.info(msg=f'requesting {tl_schema_name} with provided data {data}')
        data, qid = self.serialize_adnl_ls_query(schema, data)
        return await self.liteserver_query(data, qid)

    @staticmethod
    def pack_block_id_ext(**kwargs):
        if not kwargs.get('shard'):
            kwargs['shard'] = -9223372036854775808
        if isinstance(kwargs['root_hash'], bytes):
            kwargs['root_hash'] = kwargs['root_hash'].hex()
        if isinstance(kwargs['file_hash'], bytes):
            kwargs['file_hash'] = kwargs['file_hash'].hex()
        return {'id': {'workchain': kwargs['wc'], 'shard': kwargs['shard'], 'seqno': kwargs['seqno'], 'root_hash': kwargs['root_hash'], 'file_hash': kwargs['file_hash']}}

    async def get_trusted_last_mc_block(self):
        last_block = BlockIdExt.from_dict((await self.get_masterchain_info())['last'])
        if self.trust_level:
            return last_block
        if not self.last_key_block:
            await sync(client=self, init_block=self.init_key_block, to_block=last_block)
        await self.get_mc_block_proof(known_block=self.last_key_block, target_block=last_block)
        return last_block

    async def update_last_blocks(self):
        self.last_mc_block = await self.get_trusted_last_mc_block()
        shards = await self.raw_get_all_shards_info(self.last_mc_block)
        shard_result = {}
        for k, v in shards.items():
            shard: ShardDescr = v.list[0]
            shard_result[k] = BlockIdExt(workchain=k, seqno=shard.seq_no, shard=None, root_hash=shard.root_hash,
                                         file_hash=shard.file_hash)
        self.last_shard_blocks = shard_result
        self.logger.debug(msg=f'update blocks:\nlast_mc_block: {self.last_mc_block}\nlast_shard_blocks: {self.last_shard_blocks}')

    async def block_updater(self):
        if self.last_mc_block is None:
            self.last_mc_block = await self.get_trusted_last_mc_block()
        while True:
            await self.wait_masterchain_seqno(self.last_mc_block.seqno + 1, timeout_ms=10000, schema_name='getMasterchainInfo', data={})
            await self.update_last_blocks()

    async def get_masterchain_info(self):
        return await self.liteserver_request('getMasterchainInfo', {})

    async def raw_wait_masterchain_seqno(self, seqno: int, timeout_ms: int, suffix: bytes = b''):
        prefix = self.schemas.serialize(schema=self.schemas.get_by_name('liteServer.waitMasterchainSeqno'), data={'seqno': seqno, 'timeout_ms': timeout_ms})

        qid = get_random(32)
        data = self.schemas.serialize(
            self.adnl_query_sch,
            {'query_id': qid,
             'query': self.schemas.serialize(self.ls_query_sch,
                                             {'data': prefix + suffix}
                                             )
             }
        )
        return await self.liteserver_query(data, qid[::-1].hex())

    async def wait_masterchain_seqno(self, seqno: int, timeout_ms: int, schema_name: str, data: dict = None):
        if data is None:
            data = {}

        suffix = self.schemas.serialize(self.schemas.get_by_name('liteServer.' + schema_name), data)

        return await self.raw_wait_masterchain_seqno(seqno, timeout_ms, suffix)

    async def get_masterchain_info_ext(self):
        return await self.liteserver_request('getMasterchainInfoExt', {'mode': 0})

    async def get_time(self):
        return await self.liteserver_request('getTime', {})

    async def get_version(self):
        return await self.liteserver_request('getVersion', {})

    async def get_state(self, wc: int, shard: typing.Optional[int], seqno: int, root_hash: typing.Union[str, bytes], file_hash: typing.Union[str, bytes]):
        # TODO doesnt work: {'code': -400, 'message': 'cannot request total state: possibly too large'}
        block = self.pack_block_id_ext(wc=wc, shard=shard, seqno=seqno, root_hash=root_hash, file_hash=file_hash)
        return await self.liteserver_request('getState', block)

    async def raw_get_block_header(self, block: BlockIdExt):
        result = await self.liteserver_request('getBlockHeader', {'id': block.to_dict()} | {'mode': 0})
        h_proof = Cell.one_from_boc(result['header_proof'])
        block_id = BlockIdExt.from_dict(result['id'])
        if self.trust_level <= 1:
            check_block_header_proof(h_proof[0], block_id.root_hash)
        if not self.trust_level:
            if block_id.workchain != -1:
                await self.get_shard_block_proof(block_id)
            else:
                await self.get_mc_block_proof(known_block=self.last_key_block, target_block=block_id)
        return Block.deserialize(h_proof[0].begin_parse())

    async def get_block_header(self, wc: int, shard: typing.Optional[int], seqno: int, root_hash: typing.Union[str, bytes], file_hash: typing.Union[str, bytes]):
        block = self.pack_block_id_ext(wc=wc, shard=shard, seqno=seqno, root_hash=root_hash, file_hash=file_hash)
        return await self.raw_get_block_header(BlockIdExt.from_dict(block))

    async def lookup_block(self, wc: int, shard: int, seqno: int = -1,
                           lt: typing.Optional[int] = None, utime: typing.Optional[int] = None) -> typing.Tuple[BlockIdExt, Block]:
        """
        :param wc: block workchain
        :param shard: block shard
        :param seqno: block seqno
        :param lt: block lt
        :param utime: block unix time
        :return: tuple[blockIdExt: dict, block: Block] (block here contains only BlockInfo)
        """
        mode = 0
        if seqno != -1:
            mode = 1
        if lt is not None:
            mode = 2
        if utime is not None:
            mode = 4

        data = {'mode': mode, 'id': {'workchain': wc, 'shard': shard, 'seqno': seqno}, 'lt': lt, 'utime': utime}

        result = await self.liteserver_request('lookupBlock', data)
        h_proof = Cell.one_from_boc(result['header_proof'])
        block_id = BlockIdExt.from_dict(result['id'])
        if self.trust_level <= 1:
            check_block_header_proof(h_proof[0], block_id.root_hash)
            if not self.trust_level:
                if block_id.workchain != -1:
                    await self.get_shard_block_proof(block_id)
                else:
                    await self.get_mc_block_proof(known_block=self.last_key_block, target_block=block_id)

        return BlockIdExt.from_dict(result['id']), Block.deserialize(h_proof[0].begin_parse())

    async def raw_get_block(self, block: BlockIdExt) -> Block:
        result = await self.liteserver_request('getBlock', {'id': block.to_dict()})
        block_id = BlockIdExt.from_dict(result['id'])
        assert block_id == block
        result_block = Cell.one_from_boc(result['data'])
        if self.trust_level <= 1:
            check_block_header_proof(result_block, block_hash=block.root_hash)
            if not self.trust_level:
                await self.prove_block(block_id)
        return Block.deserialize(result_block.begin_parse())

    async def get_block(self, wc: int, shard: typing.Optional[int], seqno: int, root_hash: typing.Union[str, bytes], file_hash: typing.Union[str, bytes]):
        block = self.pack_block_id_ext(wc=wc, shard=shard, seqno=seqno, root_hash=root_hash, file_hash=file_hash)
        return await self.raw_get_block(BlockIdExt.from_dict(block))

    async def raw_get_account_state(self, address: typing.Union[str, Address], block: typing.Optional[BlockIdExt] = None) -> typing.Tuple[typing.Optional[Account], typing.Optional[ShardAccount]]:
        trusted = False
        if block is None or block == self.last_mc_block:
            block = self.last_mc_block
            trusted = True
        if isinstance(address, str):
            address = Address(address)
        account = address.to_tl_account_id()

        data = {'id': block.to_dict(), 'account': account}
        result = await self.liteserver_request('getAccountState', data)
        shrd_blk = BlockIdExt.from_dict(result['shardblk'])
        if not result['state']:
            return None, None  # account_none$0 = Account;

        account_state_root = Cell.one_from_boc(result['state'])

        if self.trust_level <= 1:
            check_shard_proof(shard_proof=result['shard_proof'], blk=block, shrd_blk=shrd_blk)
            if not trusted and not self.trust_level:
                await self.get_mc_block_proof(known_block=self.last_key_block, target_block=block)
        shard_account = check_account_proof(proof=result['proof'], shrd_blk=shrd_blk, address=address, account_state_root=account_state_root, return_account_descr=True)

        return Account.deserialize(account_state_root.begin_parse()), shard_account

    async def get_account_state(self, address: typing.Union[str, Address]) -> SimpleAccount:
        """
        :param address: account address
        :return: always SimpleAccount, even if raw_get_account_state returned None (account does not exist)
        """
        if isinstance(address, str):
            address = Address(address)
        return SimpleAccount.from_raw((await self.raw_get_account_state(address))[0], address)

    async def run_get_method(self, address: typing.Union[Address, str], method: typing.Union[int, str], stack: list) -> list:
        mode = 7  # 111
        block = self.last_mc_block

        if isinstance(address, str):
            address = Address(address)
        account = address.to_tl_account_id()

        if isinstance(method, str):
            method_id = (int.from_bytes(crc16(method.encode()), byteorder='big') & 0xffff) | 0x10000
        elif isinstance(method, int):
            method_id = method
        else:
            raise LiteClientError('provided method in unknown form')

        if isinstance(stack, list):
            stack = VmStack.serialize(stack)
        else:
            raise LiteClientError('provided stack in unknown form')

        data = {'mode': mode, 'id': block.to_dict(), 'account': account, 'method_id': method_id, 'params': stack.to_boc()}
        result = await self.liteserver_request('runSmcMethod', data)

        if result['exit_code'] != 0:
            raise RunGetMethodError(address=address, method=method, exit_code=result['exit_code'])
        if self.trust_level <= 1:
            shrd_blk = BlockIdExt.from_dict(result['shardblk'])
            check_shard_proof(shard_proof=result['shard_proof'], blk=block, shrd_blk=shrd_blk)

        return VmStack.deserialize(Slice.one_from_boc(result['result']))

    async def raw_get_shard_info(self, block: typing.Optional[BlockIdExt] = None, wc: int = 0, shard: int = -9223372036854775808, exact: bool = True):
        trusted = False
        if block is None or block == self.last_mc_block:
            block = self.last_mc_block
            trusted = True

        data = {'id': block.to_dict(), 'workchain': wc, 'shard': shard, 'exact': exact}

        result = await self.liteserver_request('getShardInfo', data)
        bin_tree = BinTree.deserialize(Slice.one_from_boc(result['shard_descr']))
        shard_descr = ShardDescr.deserialize(bin_tree.list[0])
        shrd_blk = BlockIdExt.from_dict(result['shardblk'])
        if self.trust_level <= 1:
            shard_descr_proved = check_shard_proof(shard_proof=result['shard_proof'], blk=block, shrd_blk=shrd_blk)
            if not trusted and not self.trust_level:
                await self.get_mc_block_proof(known_block=self.last_key_block, target_block=block)
            if shard_descr_proved is None:
                assert shard_descr_proved == shard_descr

        return shard_descr

    async def raw_get_all_shards_info(self, block: typing.Optional[BlockIdExt] = None) -> typing.Dict[int, BinTree]:
        """
        :param block: blockIdExt
        :return: dict[workchain: BinTree[ShardDescr]]
        """
        trusted = False
        if block is None or block == self.last_mc_block:
            block = self.last_mc_block
            trusted = True

        data = {'id': block.to_dict()}

        result = await self.liteserver_request('getAllShardsInfo', data)

        shard_hashes_cell = Cell.one_from_boc(result['data'])

        if self.trust_level <= 1:

            if not trusted and not self.trust_level:
                await self.get_mc_block_proof(known_block=self.last_key_block, target_block=block)

            proof_cells = Cell.from_boc(result['proof'])

            state_hash = check_block_header_proof(proof_cells[0][0], block_hash=block.root_hash, store_state_hash=True)

            check_proof(proof_cells[1], state_hash)

            shard_state = ShardStateUnsplit.deserialize(proof_cells[1][0].begin_parse())

            assert shard_state.shard_id.workchain_id == block.workchain
            assert shard_state.seq_no == block.seqno

            assert shard_hashes_cell[0].get_hash(0) == proof_cells[1][0][3][0].get_hash(0)  # masterchain_state_extra -> shard_hashes

        return deserialize_shard_hashes(shard_hashes_cell.begin_parse())

    async def get_all_shards_info(self, block: typing.Optional[BlockIdExt] = None) -> typing.List[BlockIdExt]:
        """
        High-level function to get shards as `BlockIdExt` list instead of BinTrees
        :param block: block to get shards of
        :return: list of BlockIdExt of all block shards
        """
        shards = await self.raw_get_all_shards_info(block)
        result = []
        for k, v in shards.items():
            for sh in v.list:
                sh: ShardDescr
                result.append(
                    BlockIdExt(
                        workchain=k,
                        shard=sh.next_validator_shard_signed,
                        seqno=sh.seq_no,
                        root_hash=sh.root_hash,
                        file_hash=sh.file_hash
                    )
                )
        return result

    async def get_one_transaction(self, address: typing.Union[Address, str], lt: int, block: BlockIdExt) -> typing.Optional[Transaction]:
        if isinstance(address, str):
            address = Address(address)

        data = {'id': block.to_dict(), 'account': address.to_tl_account_id(), 'lt': lt}
        result = await self.liteserver_request('getOneTransaction', data)
        if not result['transaction']:
            return None

        transaction_root = Cell.one_from_boc(result['transaction'])
        if self.trust_level <= 1:
            proof = Cell.one_from_boc(result['proof'])
            check_block_header_proof(proof[0], block.root_hash)

            acc_block = Block.deserialize(proof[0].begin_parse()).extra.account_blocks[0].get(int.from_bytes(address.hash_part, 'big'))
            if not acc_block:
                raise LiteClientError(f'Proof check failed! Cannot find account in account_blocks')

            acc_block: AccountBlock
            assert acc_block.account_addr == address.hash_part.hex()

            tr = acc_block.transactions[0].get(lt)
            if not tr:
                raise LiteClientError(f'Proof check failed! Cannot find transaction in account block')

            if tr.get_hash(0) != transaction_root.get_hash(0):
                raise LiteClientError(f'Proof check failed! Transaction hashes mismatch')

        return Transaction.deserialize(transaction_root.begin_parse())

    async def raw_get_transactions(self, address: typing.Union[Address, str], count: int, from_lt: int = None, from_hash: typing.Optional[bytes] = None) -> typing.Tuple[typing.List[Transaction], typing.List[BlockIdExt]]:
        if isinstance(address, str):
            address = Address(address)

        assert count <= 16, 'maximum transactions in .raw_get_transactions() is 16!'

        if not from_lt or not from_hash:
            state, shard_account = await self.raw_get_account_state(address)
            from_lt, from_hash = shard_account.last_trans_lt, shard_account.last_trans_hash

        data = {'count': count, 'account': address.to_tl_account_id(), 'lt': from_lt, 'hash': from_hash.hex()}

        result = await self.liteserver_request('getTransactions', data)

        transactions_cells = Cell.from_boc(result['transactions'])

        prev_tr_hash = from_hash

        tr_result = []
        block_ids = []
        i = 0
        for tr in transactions_cells:
            block_ids.append(BlockIdExt.from_dict(result['ids'][i]))
            if self.trust_level <= 1:
                current_hash = tr.get_hash(0)
                if current_hash != prev_tr_hash:
                    raise LiteClientError(f'Transaction hashes mismatch. Expected {prev_tr_hash}, got {current_hash}')
            transaction = Transaction.deserialize(tr.begin_parse())
            prev_tr_hash = transaction.prev_trans_hash
            tr_result.append(transaction)
            i += 1

        # assert len(tr_result) == count, f'expected {count} transactions, got {len(tr_result)}'
        return tr_result, block_ids

    async def get_transactions(self, address: typing.Union[Address, str], count: int, from_lt: int = None, from_hash: typing.Optional[bytes] = None) -> typing.List[Transaction]:
        result: typing.List[Transaction] = []

        for i in range(0, count, 16):
            amount = min(16, count - i)
            tr_result, block_ids = await self.raw_get_transactions(address, amount, from_lt, from_hash)
            result += tr_result
            from_lt, from_hash = result[-1].prev_trans_lt, result[-1].prev_trans_hash
            if from_lt == 0:
                break
        # assert len(result) == count, f'expected {count} transactions, got {len(result)}'
        return result

    async def raw_get_block_transactions(self, block: BlockIdExt, count: int = 256) -> typing.List[dict]:
        mode = 39  # 100111
        data = {'id': block.to_dict(), 'mode': mode, 'count': count, 'want_proof': b''}
        result = await self.liteserver_request('listBlockTransactions', data)
        if not result['ids']:
            return []
        transactions_ids = result['ids']
        if self.trust_level <= 1:
            proof = Cell.one_from_boc(result['proof'])
            check_block_header_proof(proof[0], block.root_hash)
            acc_block = Block.deserialize(proof[0].begin_parse()).extra.account_blocks[0]

        if not self.trust_level and block != self.last_mc_block:
            await self.prove_block(block)

        for tr in transactions_ids:
            tr['hash'] = bytes.fromhex(tr['hash'])
            tr.pop('mode')  # in this lib mode is a fixed num, so we don't really need it in result, moreover mode can mislead
            if self.trust_level <= 1:
                block_trs: dict = acc_block.get(int(tr['account'], 16)).transactions[0]
                block_tr: Cell = block_trs.get(tr['lt'])
                assert block_tr.get_hash(0) == tr['hash']
            tr['account'] = Address((block.workchain, bytes.fromhex(tr['account'])))

        return transactions_ids

    async def raw_get_block_transactions_ext(self, block: BlockIdExt, count: int = 256) -> typing.List[Transaction]:
        mode = 39  # 100111
        data = {'id': block.to_dict(), 'mode': mode, 'count': count, 'want_proof': b''}
        result = await self.liteserver_request('listBlockTransactionsExt', data)

        if not result['transactions']:
            return []

        transactions_cells = Cell.from_boc(result['transactions'])

        if self.trust_level <= 1:
            proof = Cell.one_from_boc(result['proof'])
            check_block_header_proof(proof[0], block.root_hash)
            acc_block = Block.deserialize(proof[0].begin_parse()).extra.account_blocks[0]
            if not self.trust_level and block != self.last_mc_block:
                await self.prove_block(block)
        tr_result = []

        for tr_root in transactions_cells:
            transaction = Transaction.deserialize(tr_root.begin_parse())
            if self.trust_level <= 1:
                prunned_tr_cell = acc_block.get(int(transaction.account_addr_hex, 16)).transactions[0].get(transaction.lt)
                assert prunned_tr_cell.get_hash(0) == tr_root.get_hash(0)
            # transaction.account = Address((block.workchain, bytes.fromhex(transaction.account_addr)))
            tr_result.append(transaction)

        return tr_result

    async def raw_get_mc_block_proof(self, known_block: BlockIdExt, target_block: typing.Optional[BlockIdExt] = None, return_best_key_block=False) -> typing.Tuple[bool, BlockIdExt, typing.Optional[BlockIdExt], typing.Optional[int]]:
        """
        :param known_block: block you trust
        :param target_block: block you want to prove
        :param return_best_key_block: if true the key block with big ttl will be returned
        :return: (bool, BlockIdExt, BlockIdExt, int) - is completed proof, last trusted block, best key block (see documentation), and best key block gen_utime
        """
        mode = 0

        if target_block:
            mode = 1  # 1
        data = {'known_block': known_block.to_dict(), 'mode': mode, 'target_block': target_block.to_dict()}
        result = await self.liteserver_request('getBlockProof', data)
        last_trusted = known_block
        best_key = None
        best_key_ts = 0
        for step in result['steps']:
            if 'config_proof' in step:  # blockLinkForward
                assert last_trusted == BlockIdExt.from_dict(step['from'])
                to_block = BlockIdExt.from_dict(step['to'])

                dest_proof = Cell.one_from_boc(step['dest_proof'])
                config_proof = Cell.one_from_boc(step['config_proof'])
                check_block_header_proof(dest_proof[0], to_block.root_hash)

                block = Block.deserialize(config_proof[0].begin_parse())
                dest_block = Block.deserialize(dest_proof[0].begin_parse())
                if self.last_key_block is None or block.info.seqno > self.last_key_block.seqno:
                    self.last_key_block = last_trusted
                if step['to_key_block']:
                    if self.last_key_block is None or dest_block.info.seqno > self.last_key_block.seqno:
                        self.last_key_block = to_block
                if return_best_key_block:
                    best_key, best_key_ts = choose_key_block(best_key, best_key_ts, last_trusted, block.info.gen_utime)
                    if step['to_key_block']:
                        best_key, best_key_ts = choose_key_block(best_key, best_key_ts, to_block, dest_block.info.gen_utime)

                param_34 = ConfigParam34.deserialize(block.extra.custom.config.config[34])
                param_28 = ConfigParam28.deserialize(block.extra.custom.config.config[28])

                nodes = compute_validator_set(param_28, to_block, param_34.cur_validators)
                check_block_signatures(nodes=nodes, signatures=step['signatures']['signatures'], blk=to_block)
                last_trusted = to_block

            else:  # blockLinkBack
                assert last_trusted == BlockIdExt.from_dict(step['from'])
                to_block = BlockIdExt.from_dict(step['to'])
                if step['to_key_block']:
                    dest_proof = Cell.one_from_boc(step['dest_proof'])
                    state_proof = Cell.one_from_boc(step['state_proof'])
                    proof = Cell.one_from_boc(step['proof'])
                    block = Block.deserialize(proof[0].begin_parse())
                    if return_best_key_block:
                        best_key, best_key_ts = choose_key_block(best_key, best_key_ts, to_block, block.info.gen_utime)
                    state_hash = check_block_header_proof(proof[0], last_trusted.root_hash, True)
                    assert state_hash == state_proof[0].get_hash(0)

                    state = ShardStateUnsplit.deserialize(state_proof[0].begin_parse())
                    last_key = state.custom.last_key_block
                    check_block_header_proof(dest_proof[0], last_key.root_hash)
                    assert to_block.root_hash == last_key.root_hash
                    if self.last_key_block is None or to_block.seqno > self.last_key_block.seqno:
                        self.last_key_block = to_block
                    last_trusted = to_block
                else:
                    dest_proof = Cell.one_from_boc(step['dest_proof'])  # ?
                    state_proof = Cell.one_from_boc(step['state_proof'])
                    proof = Cell.one_from_boc(step['proof'])

                    state_hash = check_block_header_proof(proof[0], last_trusted.root_hash, True)
                    assert state_hash == state_proof[0].get_hash(0)
                    state = ShardStateUnsplit.deserialize(state_proof[0].begin_parse())
                    blk = state.custom.prev_blocks[0].get(to_block.seqno)
                    if not blk:
                        raise LiteClientError(f'cannot find {to_block} in OldMcBlocksInfo')
                    blk: KeyExtBlkRef

                    assert blk.blk_ref.root_hash == to_block.root_hash
                    last_trusted = to_block
        return last_trusted == target_block, last_trusted, best_key, best_key_ts

    async def get_mc_block_proof(self, known_block: BlockIdExt, target_block: BlockIdExt, return_best_key_block=False):
        self.logger.debug(msg=f'PROOF BLOCKS\nfrom: {known_block}\ntarget: {target_block}')
        last_proved = known_block
        best_key = None
        best_key_ts = 0
        while last_proved != target_block:
            _, last_proved, key, key_ts = await self.raw_get_mc_block_proof(last_proved, target_block, return_best_key_block)
            if return_best_key_block:
                best_key, best_key_ts = choose_key_block(best_key, best_key_ts, key, key_ts)
            self.logger.debug(msg=f'PROOF BLOCKS\nproved: {last_proved}')
        if return_best_key_block:
            return best_key, best_key_ts

    async def prove_block(self, target_block: BlockIdExt):
        if target_block.workchain == -1:
            await self.get_mc_block_proof(self.last_key_block, target_block)
        else:
            await self.get_shard_block_proof(target_block, True)

    def unpack_config(self, block: BlockIdExt, config_proof: Cell, state_proof: Cell):
        if self.trust_level <= 1:
            state_hash = check_block_header_proof(state_proof[0], block.root_hash, True)
            if config_proof[0].get_hash(0) != state_hash:
                raise LiteClientError('hashes mismach')
        shard = ShardStateUnsplit.deserialize(config_proof[0].begin_parse())
        config = shard.custom.config.config
        config_res = {}
        for i, v in config.items():
            if i in ConfigParam.params:
                config_res[i] = ConfigParam.params[i].deserialize(v)
            else:
                config_res[i] = v

        return config_res

    async def get_config_all(self, blk: typing.Optional[BlockIdExt] = None):
        trusted = False
        if blk is None:
            blk = self.last_mc_block
            trusted = True

        mode = 0  # ?

        data = {'mode': mode, 'id': blk.to_dict()}

        result = await self.liteserver_request('getConfigAll', data)

        if not self.trust_level and not trusted:
            await self.prove_block(blk)

        config_proof = Cell.one_from_boc(result['config_proof'])
        state_proof = Cell.one_from_boc(result['state_proof'])
        return self.unpack_config(blk, config_proof, state_proof)

    async def get_config_params(self, params: typing.List[int], blk: typing.Optional[BlockIdExt] = None):
        trusted = False
        if blk is None:
            blk = self.last_mc_block
            trusted = True

        mode = 0  # ?
        data = {'mode': mode, 'id': blk.to_dict(), 'param_list': params}
        result = await self.liteserver_request('getConfigParams', data)

        if not self.trust_level and not trusted:
            await self.prove_block(blk)

        config_proof = Cell.one_from_boc(result['config_proof'])
        state_proof = Cell.one_from_boc(result['state_proof'])
        return self.unpack_config(blk, config_proof, state_proof)

    async def get_libraries(self, library_list: typing.List[bytes]):
        data = {'library_list': library_list}

        result = await self.liteserver_request('getLibraries', data)

        return result['result']

    async def get_shard_block_proof(self, blk: BlockIdExt, prove_mc: bool = False):
        data = {'id': blk.to_dict()}

        result = await self.liteserver_request('getShardBlockProof', data)
        mc_block = BlockIdExt.from_dict(result['masterchain_id'])

        if prove_mc:
            await self.get_mc_block_proof(known_block=self.last_key_block, target_block=mc_block)

        def check_shard_in_master(proof: Cell, blk: BlockIdExt):
            check_block_header_proof(proof[0], mc_block.root_hash)

            shards = Block.deserialize(proof[0].begin_parse()).extra.custom.shard_hashes[blk.workchain].list
            shard = None
            for sh in shards:
                sh: ShardDescr
                if sh.seq_no == blk.seqno and sh.next_validator_shard_signed == blk.shard:
                    shard = sh.__dict__

            shardblk = BlockIdExt.from_dict(shard)
            shardblk.seqno = shard['seq_no']
            shardblk.workchain = blk.workchain
            return shardblk

        if len(result['links']) == 1:
            assert check_shard_in_master(Cell.one_from_boc(result['links'][0]['proof']), blk) == blk
            return

        last_shard_blk = None

        for link in result['links']:
            proof = Cell.one_from_boc(link['proof'])
            if proof[0].get_hash(0) == mc_block.root_hash:
                last_shard_blk = check_shard_in_master(proof, BlockIdExt.from_dict(link['id']))
                continue
            check_block_header_proof(proof[0], last_shard_blk.root_hash)
            shrd_blk = Block.deserialize(proof[0].begin_parse())
            prev_blk = shrd_blk.info.prev_ref.prev

            last_shard_blk = BlockIdExt.from_dict(prev_blk.__dict__ | {'workchain': last_shard_blk.workchain, 'shard': last_shard_blk.shard})
            if last_shard_blk == blk:
                return
            raise LiteClientError('incorrect proof')

    async def raw_send_message(self, message: bytes):
        data = {'body': message}

        result = await self.liteserver_request('sendMessage', data)
        return result['status']

    @classmethod
    def from_config(cls, config: dict, ls_i: int = 0, trust_level: int = 2, timeout: int = 10):
        ls = config['liteservers'][ls_i]
        init_block = config['validator']['init_block'].copy()
        init_block['file_hash'] = base64.b64decode(init_block['file_hash']).hex()
        init_block['root_hash'] = base64.b64decode(init_block['root_hash']).hex()
        init_block = BlockIdExt.from_dict(init_block)
        if not trust_level and init_block != init_mainnet_block and init_block != init_testnet_block:
            logging.getLogger(cls.__name__).warning(msg='unknown init block found! please, check its hash to trust it')

        return cls(
            host=socket.inet_ntoa(struct.pack('>i', ls['ip'])),
            port=ls['port'],
            server_pub_key=ls['id']['key'],
            trust_level=trust_level,
            init_key_block=init_block
        )

    @classmethod
    def from_mainnet_config(cls, ls_i: int = 0, trust_level: int = 0, timeout: int = 10):
        config = requests.get('https://ton.org/global-config.json').json()
        return cls.from_config(config, ls_i, trust_level, timeout)

    @classmethod
    def from_testnet_config(cls, ls_i: int = 0, trust_level: int = 0, timeout: int = 10):
        config = requests.get('https://ton.org/testnet-global.config.json').json()
        return cls.from_config(config, ls_i, trust_level, timeout)



import asyncio
import logging

import pytest

import random
from pytoniq import LiteClient


@pytest.mark.asyncio
async def test_init():
    client = LiteClient.from_mainnet_config(random.randint(0, 8), trust_level=2)
    await client.connect()
    await client.close()

    client = LiteClient.from_testnet_config(random.randint(0, 5), trust_level=2)
    await client.connect()
    await client.close()

    # try:
    #     client = LiteClient.from_mainnet_config(random.randint(0, 8), trust_level=0)
    #     await client.connect()
    #     await client.close()
    # except asyncio.TimeoutError:
    #     print('skipping')


@pytest.mark.asyncio
async def test_methods():
    while True:
        client = LiteClient.from_mainnet_config(random.randint(0, 8), trust_level=1)
        try:
            await client.connect()
            break
        except asyncio.TimeoutError:
            await client.close()
            continue
    await client.get_masterchain_info()
    await client.get_config_all()
    await client.raw_get_block(client.last_mc_block)
    await client.close()


@pytest.mark.asyncio
async def test_get_method():
    while True:
        client = LiteClient.from_mainnet_config(random.randint(0, 8), trust_level=1)
        try:
            await client.connect()
            break
        except asyncio.TimeoutError:
            await client.close()
            continue

    result = await client.run_get_method(address='EQBvW8Z5huBkMJYdnfAEM5JqTNkuWX3diqYENkWsIL0XggGG', method='seqno',
                                         stack=[])
    assert isinstance(result[0], int)



import time
import typing

from .wallet import Wallet, WalletError
from ..utils import generate_query_id
from ..contract import Contract, ContractError
from ...liteclient.client import LiteClient
from pytoniq_core.crypto.keys import private_key_to_public_key, mnemonic_to_private_key, mnemonic_is_valid, mnemonic_new
from pytoniq_core.crypto.signature import sign_message
from pytoniq_core.boc import Cell, Builder, HashMap
from pytoniq_core.boc.address import Address
from pytoniq_core.tlb.account import StateInit
from pytoniq_core.tlb.custom.wallet import HighloadWalletData, WalletMessage

HIGHLOAD_WALLET_CODE = Cell.one_from_boc(
    b'\xb5\xee\x9cr\x01\x01\t\x01\x00\xe5\x00\x01\x14\xff\x00\xf4\xa4\x13\xf4\xbc\xf2\xc8\x0b\x01\x02\x01 \x02\x03\x02\x01H\x04\x05\x01\xea\xf2\x83\x08\xd7\x18 \xd3\x1f\xd3?\xf8#\xaa\x1fS \xb9\xf2c\xedD\xd0\xd3\x1f\xd3?\xd3\xff\xf4\x04\xd1S`\x80@\xf4\x0eo\xa11\xf2`Qs\xba\xf2\xa2\x07\xf9\x01T\x10\x87\xf9\x10\xf2\xa3\x02\xf4\x04\xd1\xf8\x00\x7f\x8e\x16!\x80\x10\xf4xo\xa5 \x98\x02\xd3\x07\xd40\x01\xfb\x00\x912\xe2\x01\xb3\xe6[\x83%\xa1\xc8@4\x80@\xf4C\x8a\xe61\x01\xc8\xcb\x1f\x13\xcb?\xcb\xff\xf4\x00\xc9\xedT\x08\x00\x04\xd00\x02\x01 \x06\x07\x00\x17\xbd\x9c\xe7j&\x86\x9a\xf9\x8e\xb8_\xfc\x00A\xbe_\x97j&\x86\x98\xf9\x8e\x99\xfe\x9f\xf9\x8f\xa0&\x8a\x91\x04\x02\x07\xa0s}\t\x8c\x92\xdb\xfc\x95\xdd\x1f\x14\x004 \x80@\xf4\x96o\xa5l\x12 \x940S\x03\xb9\xde \x9336\x01\x92l!\xe2\xb3')


class HighloadWallet(Wallet):

    @classmethod
    async def from_data(cls, provider: LiteClient, public_key: bytes, wc: int = 0,
                        wallet_id: typing.Optional[int] = None, **kwargs) -> "HighloadWallet":
        data = cls.create_data_cell(public_key, wallet_id, wc)
        return await super().from_code_and_data(provider, wc, HIGHLOAD_WALLET_CODE, data, **kwargs)

    @staticmethod
    def create_data_cell(public_key: bytes, wallet_id: typing.Optional[int] = None, wc: typing.Optional[int] = 0,
                         old_queries: typing.Optional[dict] = None) -> Cell:
        if wallet_id is None:
            wallet_id = 698983191 + wc
        return HighloadWalletData(wallet_id=wallet_id, public_key=public_key, last_cleaned=0, old_queries=old_queries).serialize()

    @classmethod
    async def from_private_key(cls, provider: LiteClient, private_key: bytes, wc: int = 0,
                               wallet_id: typing.Optional[int] = None):
        public_key = private_key_to_public_key(private_key)
        return await cls.from_data(provider=provider, wc=wc, public_key=public_key, wallet_id=wallet_id,
                                   private_key=private_key)

    @classmethod
    async def from_mnemonic(cls, provider: LiteClient, mnemonics: typing.Union[list, str], wc: int = 0,
                            wallet_id: typing.Optional[int] = None):
        if isinstance(mnemonics, str):
            mnemonics = mnemonics.split()
        assert mnemonic_is_valid(mnemonics), 'mnemonics are invalid!'
        _, private_key = mnemonic_to_private_key(mnemonics)
        return await cls.from_private_key(provider, private_key, wc, wallet_id)

    @classmethod
    async def create(cls, provider: LiteClient, wc: int = 0, wallet_id: typing.Optional[int] = None):
        """
        :param provider: provider
        :param wc: wallet workchain
        :param wallet_id: subwallet_id
        :return: mnemonics and Wallet instance of provided version
        """
        mnemo = mnemonic_new(24)
        return mnemo, await cls.from_mnemonic(provider, mnemo, wc, wallet_id)

    @staticmethod
    def raw_create_transfer_msg(private_key: bytes, wallet_id: int, messages: typing.List[WalletMessage],
                                query_id: int = 0, offset: int = 7200) -> Cell:

        signing_message = Builder().store_uint(wallet_id, 32)
        if not query_id:
            signing_message.store_uint(generate_query_id(offset), 64)
        else:
            signing_message.store_uint(query_id, 64)

        def value_serializer(src, dest):
            dest.store_cell(src.serialize())

        messages_dict = HashMap(key_size=16, value_serializer=value_serializer)

        for i in range(len(messages)):
            messages_dict.set_int_key(i, messages[i])

        signing_message.store_dict(messages_dict.serialize())

        signing_message = signing_message.end_cell()
        signature = sign_message(signing_message.hash, private_key)
        return Builder() \
            .store_bytes(signature) \
            .store_cell(signing_message) \
            .end_cell()

    async def raw_transfer(self, msgs: typing.List[WalletMessage], query_id: int = 0, offset: int = 7200):
        """
        :param query_id: query id
        :param offset: if query id is 0 it will be generated as current_time + offset
        :param msgs: list of WalletMessages. to create one call create_wallet_internal_message meth
        """
        assert len(msgs) <= 254, 'for highload wallet maximum messages amount is 254'
        if 'private_key' not in self.__dict__:
            raise WalletError('must specify wallet private key!')

        transfer_msg = self.raw_create_transfer_msg(private_key=self.private_key, wallet_id=self.wallet_id,
                                                    query_id=query_id, offset=offset, messages=msgs)

        return await self.send_external(body=transfer_msg)

    async def transfer(self, destinations: typing.Union[Address, str], amounts: int, bodies: typing.List[Cell],
                       state_inits: typing.List[StateInit] = None):
        result_msgs = []
        for i in range(len(destinations)):
            destination = destinations[i]
            body = bodies[i]
            if body is None:
                body = Cell.empty()

            if isinstance(destination, str):
                destination = Address(destination)

            result_msgs.append(
                self.create_wallet_internal_message(destination=destination, value=amounts[i], body=body,
                                                    state_init=state_inits[i]))
        return await self.raw_transfer(msgs=result_msgs)

    async def send_init_external(self):
        if not self.state_init:
            raise ContractError('contract does not have state_init attribute')
        if 'private_key' not in self.__dict__:
            raise WalletError('must specify wallet private key!')
        body = self.raw_create_transfer_msg(private_key=self.private_key, wallet_id=self.wallet_id, messages=[])
        return await self.send_external(state_init=self.state_init, body=body)

    @property
    def wallet_id(self) -> int:
        """
        :return: wallet_id taken from contract data
        """
        return HighloadWalletData.deserialize(self.state.data.begin_parse()).wallet_id

    @property
    def last_cleaned(self) -> int:
        """
        :return: last_cleaned taken from contract data
        """
        return HighloadWalletData.deserialize(self.state.data.begin_parse()).last_cleaned

    @property
    def public_key(self) -> bytes:
        """
        :return: public_key taken from contract data
        """
        return HighloadWalletData.deserialize(self.state.data.begin_parse()).public_key

    @property
    def old_queries(self) -> dict:
        """
        :return: old_queries taken from contract data
        """
        return HighloadWalletData.deserialize(self.state.data.begin_parse()).old_queries

    async def processed(self, query_id: int) -> bool:
        """
        :return: is query processed from wallet's get method
        """
        return (await super().run_get_method(method='processed?', stack=[query_id]))[0]



import time
import typing

from ..contract import Contract, ContractError
from ...liteclient.client import LiteClient
from pytoniq_core.boc import Cell, Builder, HashMap
from pytoniq_core.boc.address import Address
from pytoniq_core.tlb.custom.nft import NftItemSaleData, NftItemSaleFees


# https://github.com/getgems-io/nft-contracts/blob/main/packages/contracts/sources/nft-fixprice-sale-v3r2.fc
NFT_SALE_CODE = Cell.one_from_boc(b"\xb5\xee\x9crA\x02\x0b\x01\x00\x02\xb9\x00\x01\x14\xff\x00\xf4\xa4\x13\xf4\xbc\xf2\xc8\x0b\x01\x02\x01 \x02\x03\x02\x01H\x04\x05\x00~\xf20\xedD\xd0\xd3\x00\xd3\x1f\xfa@\xfa@\xfa@\xfa\x00\xd4\xd3\x000\xc0\x01\x8e\x1d\xf8\x00p\x07\xc8\xcb\x00\x16\xcb\x1fP\x04\xcf\x16X\xcf\x16\x01\xcf\x16\x01\xfa\x02\xcc\xcb\x00\xc9\xedT\xe0_\x07\x82\x00\xff\xfe\xf2\xf0\x02\x02\xcd\x06\x07\x00W\xa08Y\xda\x89\xa1\xa6\x01\xa6?\xf4\x81\xf4\x81\xf4\x81\xf4\x01\xa9\xa6\x00`a\xa1\xf4\x81\xf4\x01\xf4\x81\xf4\x00a\x04 \x8c\x92\xb0\xa0\x15\x80\x02\xab\x01\x01\xf7\xd0\x0e\x86\x98\x18\x0b\x8d\x84\x92\xf8'\x07\xd2\x01\x87j&\x86\x98\x06\x98\xff\xd2\x07\xd2\x07\xd2\x07\xd0\x06\xa6\x98\x01\x81\x83\x82\x98N8\x06\x00\x04\xa9\x88N\x98\xf8V\xf1\x0e\x18\x04\xa1\x80N\x99\xfcp\x8c[1\xb0\xb71\xb2\xb6A^8,\x93\x99\x96\xf2\x80W\x11V\x00\x0c\x92\xf8o\x01&\xbaN\x10\x11\\\x08\x11]\xd1V\x00\t\x15\x9d\x8d\x82\x9d\xc68-\x84\xe8\xea\xf8n\xa1\x86\x86\x98>\xa1\x80\x0f\xd8\x07\x01N\x00\x0c\x08\x01\xf7f\x08@\xeek(\x01I\x82\x81H\xc2\xfb\xcb\x87\x08\x93C\xe9\x03\xe8\x03\xe9\x03\xe8\x00\xc1NJ\x84\x86\x85B\x1e\x84Z\x81JA\xc2\x00C#,\x15@\x0f<X\x07\xe8\x0b-\xab%\xc7\xec\x00\x97\x08\x00\x97]'\x08\n\xc28]A\x15\xc2\x00C#,\x15@\x0f<X\x07\xe8\x0b-\xab%\xc7\xec\x00@\x8eH\xd0\xd3\x89i\xc2\x00C#,\x15@\x0f<X\x07\xe8\x0b-\xab%\xc7\xec\x01\xc0\x82\x08A\x7f0\xf4R\n\x01\xe8\xf2\xd1\x94\xb3\x8eB1399SR\xc7\x05\x92_\t\xe0QQ\xc7\x05\xf2\xe1\xf4\x82\x10\x05\x13\x8d\x91\x16\xba\xf2\xe1\xf5\x03\xfa@0FP\x104Yp\x07\xc8\xcb\x00\x16\xcb\x1fP\x04\xcf\x16X\xcf\x16\x01\xcf\x16\x01\xfa\x02\xcc\xcb\x00\xc9\xedT\xe007(\xc0\x03\xe3\x02(\xc0\x00\x9c67\x108Ge\x14C0p\xf0\x05\xe0\x08\xc0\x02\x98UD\x10$\x10#\xf0\x05\xe0_\n\x84\x0f\xf2\xf0\t\x00\xd489\x82\x10;\x9a\xca\x00\x18\xbe\xf2\xe1\xc9SF\xc7\x05QR\xc7\x05\x15\xb1\xf2\xe1\xcap \x82\x10_\xcc=\x14!\x80\x10\xc8\xcb\x05(\xcf\x16!\xfa\x02\xcbj\xcb\x1f\x15\xcb?'\xcf\x16'\xcf\x16\x14\xca\x00#\xfa\x02\x13\xca\x00\xc9\x83\x06\xfb\x00qPfE\x15\x04p\x07\xc8\xcb\x00\x16\xcb\x1fP\x04\xcf\x16X\xcf\x16\x01\xcf\x16\x01\xfa\x02\xcc\xcb\x00\xc9\xedT\x00\x96\xc8\xcb\x1f\x13\xcb?#\xcf\x16P\x03\xcf\x16\xca\x00\x82\t\xc9\xc3\x80\xfa\x02\xca\x00\xc9q\x80\x18\xc8\xcb\x05&\xcf\x16p\xfa\x02\xcbj\xcc\xc9\x83\x06\xfb\x00qUPp\x07\xc8\xcb\x00\x16\xcb\x1fP\x04\xcf\x16X\xcf\x16\x01\xcf\x16\x01\xfa\x02\xcc\xcb\x00\xc9\xedT\xd6^e\x89")


class NftItemSale(Contract):

    @classmethod
    async def from_data(cls, provider: LiteClient, marketplace_address: Address, nft_address: Address, nft_owner_address: Address, full_price: int, fees: NftItemSaleFees, wc: int = 0, **kwargs) -> "NftItemSale":
        data = cls.create_data_cell(is_complete=False, created_at=int(time.time()), marketplace_address=marketplace_address, nft_address=nft_address, nft_owner_address=nft_owner_address, full_price=full_price, fees=fees, can_deploy_by_external=True)
        return await super().from_code_and_data(provider, wc, NFT_SALE_CODE, data, **kwargs)

    @staticmethod
    def create_data_cell(is_complete: bool, created_at: int, marketplace_address: Address, nft_address: Address, nft_owner_address: Address, full_price: int, fees: NftItemSaleFees, can_deploy_by_external: bool) -> Cell:
        return NftItemSaleData(is_complete, created_at, marketplace_address, nft_address, nft_owner_address, full_price, fees, can_deploy_by_external).serialize()

    @staticmethod
    def create_fees_cell(marketplace_fee_address: Address, marketplace_fee: int, royalty_address: Address, royalty_amount: int) -> Cell:
        return NftItemSaleFees(marketplace_fee_address, marketplace_fee, royalty_address, royalty_amount).serialize()

    @property
    def is_complete(self) -> bool:
        """
        :return: is_complete taken from contract data
        """
        return NftItemSaleData.deserialize(self.state.data.begin_parse()).is_complete

    @property
    def created_at(self) -> int:
        """
        :return: created_at taken from contract data
        """
        return NftItemSaleData.deserialize(self.state.data.begin_parse()).created_at

    @property
    def marketplace_address(self) -> Address:
        """
        :return: marketplace_address taken from contract data
        """
        return NftItemSaleData.deserialize(self.state.data.begin_parse()).marketplace_address

    @property
    def nft_address(self) -> Address:
        """
        :return: nft_address taken from contract data
        """
        return NftItemSaleData.deserialize(self.state.data.begin_parse()).nft_address

    @property
    def nft_owner_address(self) -> Address:
        """
        :return: nft_owner_address taken from contract data
        """
        return NftItemSaleData.deserialize(self.state.data.begin_parse()).nft_owner_address

    @property
    def full_price(self) -> int:
        """
        :return: full_price taken from contract data
        """
        return NftItemSaleData.deserialize(self.state.data.begin_parse()).full_price

    @property
    def fees_cell(self) -> NftItemSaleFees:
        """
        :return: fees_cell taken from contract data
        """
        return NftItemSaleData.deserialize(self.state.data.begin_parse()).fees_cell

    @property
    def can_deploy_by_external(self) -> bool:
        """
        :return: can_deploy_by_external taken from contract data
        """
        return NftItemSaleData.deserialize(self.state.data.begin_parse()).can_deploy_by_external

    async def get_sale_data(self) -> bool:
        return await super().run_get_method(method='get_sale_data', stack=[])



import time
import typing

from ...liteclient.client import LiteClient
from ..contract import Contract, ContractError
from pytoniq_core.crypto.keys import private_key_to_public_key, mnemonic_to_private_key, mnemonic_is_valid, mnemonic_new
from pytoniq_core.crypto.signature import sign_message
from pytoniq_core.boc import Cell, Builder
from pytoniq_core.boc.address import Address
from pytoniq_core.tlb.account import StateInit
from pytoniq_core.tlb.custom.wallet import WalletV3Data, WalletV4Data, WalletMessage

WALLET_V3_R2_CODE = Cell.one_from_boc(
    b'\xb5\xee\x9crA\x01\x01\x01\x00q\x00\x00\xde\xff\x00 \xdd \x82\x01L\x97\xba!\x82\x013\x9c\xba\xb1\x9fq\xb0\xedD\xd0\xd3\x1f\xd3\x1f1\xd7\x0b\xff\xe3\x04\xe0\xa4\xf2`\x83\x08\xd7\x18 \xd3\x1f\xd3\x1f\xd3\x1f\xf8#\x13\xbb\xf2c\xedD\xd0\xd3\x1f\xd3\x1f\xd3\xff\xd1Q2\xba\xf2\xa1QD\xba\xf2\xa2\x04\xf9\x01T\x10U\xf9\x10\xf2\xa3\xf8\x00\x93 \xd7J\x96\xd3\x07\xd4\x02\xfb\x00\xe8\xd1\x01\xa4\xc8\xcb\x1f\xcb\x1f\xcb\xff\xc9\xedT\x10\xbdm\xad')
WALLET_V3_R1_CODE = Cell.one_from_boc(
    b'\xb5\xee\x9crA\x01\x01\x01\x00b\x00\x00\xc0\xff\x00 \xdd \x82\x01L\x97\xba\x970\xedD\xd0\xd7\x0b\x1f\xe0\xa4\xf2`\x83\x08\xd7\x18 \xd3\x1f\xd3\x1f\xd3\x1f\xf8#\x13\xbb\xf2c\xedD\xd0\xd3\x1f\xd3\x1f\xd3\xff\xd1Q2\xba\xf2\xa1QD\xba\xf2\xa2\x04\xf9\x01T\x10U\xf9\x10\xf2\xa3\xf8\x00\x93 \xd7J\x96\xd3\x07\xd4\x02\xfb\x00\xe8\xd1\x01\xa4\xc8\xcb\x1f\xcb\x1f\xcb\xff\xc9\xedT?\xben\xe0')
WALLET_V4_R2_CODE = Cell.one_from_boc(
    b'\xb5\xee\x9crA\x02\x14\x01\x00\x02\xd4\x00\x01\x14\xff\x00\xf4\xa4\x13\xf4\xbc\xf2\xc8\x0b\x01\x02\x01 \x02\x03\x02\x01H\x04\x05\x04\xf8\xf2\x83\x08\xd7\x18 \xd3\x1f\xd3\x1f\xd3\x1f\x02\xf8#\xbb\xf2d\xedD\xd0\xd3\x1f\xd3\x1f\xd3\xff\xf4\x04\xd1QC\xba\xf2\xa1QQ\xba\xf2\xa2\x05\xf9\x01T\x10d\xf9\x10\xf2\xa3\xf8\x00$\xa4\xc8\xcb\x1fR@\xcb\x1fR0\xcb\xffR\x10\xf4\x00\xc9\xedT\xf8\x0f\x01\xd3\x07!\xc0\x00\x9flQ\x93 \xd7J\x96\xd3\x07\xd4\x02\xfb\x00\xe80\xe0!\xc0\x01\xe3\x00!\xc0\x02\xe3\x00\x01\xc0\x03\x910\xe3\r\x03\xa4\xc8\xcb\x1f\x12\xcb\x1f\xcb\xff\x10\x11\x12\x13\x02\xe6\xd0\x01\xd0\xd3\x03!q\xb0\x92_\x04\xe0"\xd7I\xc1 \x92_\x04\xe0\x02\xd3\x1f!\x82\x10plug\xbd"\x82\x10dstr\xbd\xb0\x92_\x05\xe0\x03\xfa@0 \xfaD\x01\xc8\xca\x07\xcb\xff\xc9\xd0\xedD\xd0\x81\x01@\xd7!\xf4\x040\\\x81\x01\x08\xf4\no\xa11\xb3\x92_\x07\xe0\x05\xd3?\xc8%\x82\x10plug\xba\x9280\xe3\r\x03\x82\x10dstr\xba\x92_\x06\xe3\r\x06\x07\x02\x01 \x08\t\x00x\x01\xfa\x00\xf4\x040\xf8\'o"0P\n\xa1!\xbe\xf2\xe0P\x82\x10plug\x83\x1e\xb1p\x80\x18P\x04\xcb\x05&\xcf\x16X\xfa\x02\x19\xf4\x00\xcbi\x17\xcb\x1fR`\xcb? \xc9\x80@\xfb\x00\x06\x00\x8aP\x04\x81\x01\x08\xf4Y0\xedD\xd0\x81\x01@\xd7 \xc8\x01\xcf\x16\xf4\x00\xc9\xedT\x01r\xb0\x8e#\x82\x10dstr\x83\x1e\xb1p\x80\x18P\x05\xcb\x05P\x03\xcf\x16#\xfa\x02\x13\xcbj\xcb\x1f\xcb?\xc9\x80@\xfb\x00\x92_\x03\xe2\x02\x01 \n\x0b\x00Y\xbd$+oj&\x84\x08\n\x06\xb9\x0f\xa0!\x84p\xd4\x08\x08G\xa4\x93})\x91\x0c\xe6\x90>\x9f\xf9\x83x\x12\x80\x1bx\x10\x14\x89\x87\x15\x9f1\x84\x02\x01X\x0c\r\x00\x11\xb8\xc9~\xd4M\rp\xb1\xf8\x00=\xb2\x9d\xfbQ4 @P5\xc8}\x01\x0c\x00\xb22\x81\xf2\xff\xf2t\x00`@B=\x02\x9b\xe8L`\x02\x01 \x0e\x0f\x00\x19\xad\xcev\xa2h@ k\x90\xeb\x85\xff\xc0\x00\x19\xaf\x1d\xf6\xa2h@\x10k\x90\xeb\x85\x8f\xc0\x00n\xd2\x07\xfa\x00\xd4\xd4"\xf9\x00\x05\xc8\xca\x07\x15\xcb\xff\xc9\xd0wt\x80\x18\xc8\xcb\x05\xcb\x02"\xcf\x16P\x05\xfa\x02\x14\xcbk\x12\xcc\xcc\xc9s\xfb\x00\xc8@\x14\x81\x01\x08\xf4Q\xf2\xa7\x02\x00p\x81\x01\x08\xd7\x18\xfa\x00\xd3?\xc8T G\x81\x01\x08\xf4Q\xf2\xa7\x82\x10notept\x80\x18\xc8\xcb\x05\xcb\x02P\x06\xcf\x16P\x04\xfa\x02\x14\xcbj\x12\xcb\x1f\xcb?\xc9s\xfb\x00\x02\x00l\x81\x01\x08\xd7\x18\xfa\x00\xd3?0R$\x81\x01\x08\xf4Y\xf2\xa7\x82\x10dstrpt\x80\x18\xc8\xcb\x05\xcb\x02P\x05\xcf\x16P\x03\xfa\x02\x13\xcbj\xcb\x1f\x12\xcb?\xc9s\xfb\x00\x00\n\xf4\x00\xc9\xedTib%\xe5')


class WalletError(ContractError):
    pass


class Wallet(Contract):
    @classmethod
    async def from_private_key(cls, *args, **kwargs): ...

    @staticmethod
    def raw_create_transfer_msg(*args, **kwargs): ...

    @classmethod
    async def from_mnemonic(cls, *args, **kwargs): ...

    @classmethod
    async def create(cls, *args, **kwargs): ...

    @staticmethod
    def create_wallet_internal_message(destination: Address, send_mode: int = 3, value: int = 0, body: typing.Union[Cell, str] = None,
                                       state_init: typing.Optional[StateInit] = None, **kwargs) -> WalletMessage:
        if isinstance(body, str):
            body = Builder()\
                .store_uint(0, 32)\
                .store_string(body)\
                .end_cell()

        message = Contract.create_internal_msg(dest=destination, value=value, body=body, state_init=state_init, **kwargs)
        return WalletMessage(send_mode=send_mode, message=message)

    async def send_init_external(self): ...

    async def raw_transfer(self, *args, **kwargs): ...

    async def transfer(self, *args, **kwargs): ...


class BaseWallet(Wallet):
    """
    class for user wallets such as v4r2, v3r2, etc.
    """

    @classmethod
    async def from_private_key(cls, provider: LiteClient, private_key: bytes, wc: int = 0,
                               wallet_id: typing.Optional[int] = None, version: str = 'v3r2'):
        public_key = private_key_to_public_key(private_key)
        if version == 'v3r2':
            return await WalletV3R2.from_data(provider=provider, wc=wc, public_key=public_key, wallet_id=wallet_id,
                                              private_key=private_key)
        elif version == 'v4r2':
            return await WalletV4R2.from_data(provider=provider, wc=wc, public_key=public_key, wallet_id=wallet_id,
                                              private_key=private_key)
        elif version == 'v3r1':
            return await WalletV3R1.from_data(provider=provider, wc=wc, public_key=public_key, wallet_id=wallet_id,
                                              private_key=private_key)
        else:
            raise Exception(f'Wallet version {version} does not supported')

    @classmethod
    async def from_mnemonic(cls, provider: LiteClient, mnemonics: typing.Union[list, str], wc: int = 0,
                            wallet_id: typing.Optional[int] = None, version: str = 'v3r2'):
        if isinstance(mnemonics, str):
            mnemonics = mnemonics.split()
        assert mnemonic_is_valid(mnemonics), 'mnemonics are invalid!'
        _, private_key = mnemonic_to_private_key(mnemonics)
        return await cls.from_private_key(provider, private_key, wc, wallet_id, version)

    @classmethod
    async def create(cls, provider: LiteClient, wc: int = 0, wallet_id: typing.Optional[int] = None,
                     version: str = 'v3r2'):
        """
        :param provider: provider
        :param wc: wallet workchain
        :param wallet_id: subwallet_id
        :param version: wallet version
        :return: mnemonics and Wallet instance of provided version
        """
        mnemo = mnemonic_new(24)
        return mnemo, await cls.from_mnemonic(provider, mnemo, wc, wallet_id, version)

    @staticmethod
    def raw_create_transfer_msg(private_key: bytes, seqno: int, wallet_id: int, messages: typing.List[WalletMessage],
                                valid_until: typing.Optional[int] = None) -> Cell:
        signing_message = Builder().store_uint(wallet_id, 32)
        if seqno == 0:
            signing_message.store_bits('1' * 32)  # bin(2**32 - 1)
        else:
            if valid_until is not None:
                signing_message.store_uint(valid_until, 32)
            else:
                signing_message.store_uint(int(time.time()) + 60, 32)
        signing_message.store_uint(seqno, 32)
        for m in messages:
            signing_message.store_cell(m.serialize())
        signing_message = signing_message.end_cell()
        signature = sign_message(signing_message.hash, private_key)
        return Builder() \
            .store_bytes(signature) \
            .store_cell(signing_message) \
            .end_cell()

    async def raw_transfer(self, msgs: typing.List[WalletMessage], seqno_from_get_meth: bool = True):
        """
        :param msgs: list of WalletMessages. to create one call create_wallet_internal_message meth
        :param seqno_from_get_meth: if True LiteClient will request seqno get method and use it, otherwise seqno from contract data will be taken
        """
        assert len(msgs) <= 4, 'for common wallet maximum messages amount is 4'
        if 'private_key' not in self.__dict__:
            raise WalletError('must specify wallet private key!')

        if seqno_from_get_meth:
            seqno = await self.get_seqno()
        else:
            seqno = self.seqno
        transfer_msg = self.raw_create_transfer_msg(private_key=self.private_key, seqno=seqno, wallet_id=self.wallet_id, messages=msgs)

        return await self.send_external(body=transfer_msg)

    async def transfer(self, destination: typing.Union[Address, str], amount: int, body: Cell = Cell.empty(),
                       state_init: StateInit = None):
        if isinstance(destination, str):
            destination = Address(destination)
        wallet_message = self.create_wallet_internal_message(destination=destination, value=amount, body=body, state_init=state_init)
        return await self.raw_transfer(msgs=[wallet_message])

    async def send_init_external(self):
        if not self.state_init:
            raise ContractError('contract does not have state_init attribute')
        if 'private_key' not in self.__dict__:
            raise WalletError('must specify wallet private key!')
        body = self.raw_create_transfer_msg(private_key=self.private_key, seqno=0, wallet_id=self.wallet_id, messages=[])
        return await self.send_external(state_init=self.state_init, body=body)

    async def get_seqno(self) -> int:
        """
        :return: seqno from wallet's get method
        """
        return (await super().run_get_method('seqno'))[0]

    async def get_public_key(self) -> int:
        """
        :return: public key from wallet's get method
        """
        if self.__class__ == WalletV3R1:
            raise Exception('WalletV3R1 doesn\'t have get_public_key get method. Use .public_key attribute')
        return (await super().run_get_method('get_public_key'))[0]

    async def deploy_via_internal(self, contract: Contract, deploy_amount: int = int(0.05 * 10**9)):
        return await self.transfer(destination=contract.address, amount=deploy_amount, state_init=contract.state_init)


class WalletV3(BaseWallet):

    @classmethod
    async def from_code_and_data(cls, provider: LiteClient, code: Cell, public_key: bytes, wc: int = 0,
                                 wallet_id: typing.Optional[int] = None, private_key: typing.Optional[bytes] = None):
        data = cls.create_data_cell(public_key, wallet_id, wc)
        return await super().from_code_and_data(provider, wc, code, data, private_key=private_key)

    @staticmethod
    def create_data_cell(public_key: bytes, wallet_id: typing.Optional[int] = None,
                         wc: typing.Optional[int] = 0) -> Cell:
        if wallet_id is None:
            wallet_id = 698983191 + wc
        return WalletV3Data(seqno=0, wallet_id=wallet_id, public_key=public_key).serialize()

    @property
    def seqno(self) -> int:
        """
        :return: seqno taken from contract data
        """
        return WalletV3Data.deserialize(self.state.data.begin_parse()).seqno

    @property
    def wallet_id(self) -> int:
        """
        :return: wallet_id taken from contract data
        """
        return WalletV3Data.deserialize(self.state.data.begin_parse()).wallet_id

    @property
    def public_key(self) -> bytes:
        """
        :return: public_key taken from contract data
        """
        return WalletV3Data.deserialize(self.state.data.begin_parse()).public_key


class WalletV4(BaseWallet):

    @classmethod
    async def from_code_and_data(cls, provider: LiteClient, code: Cell, public_key: bytes, wc: int = 0,
                                 wallet_id: typing.Optional[int] = None, **kwargs):
        data = cls.create_data_cell(public_key, wallet_id, wc)
        return await super().from_code_and_data(provider, wc, code, data, **kwargs)

    @staticmethod
    def create_data_cell(public_key: bytes, wallet_id: typing.Optional[int] = None, wc: typing.Optional[int] = 0,
                         plugins: typing.Optional[Cell] = None) -> Cell:
        if wallet_id is None:
            wallet_id = 698983191 + wc
        return WalletV4Data(seqno=0, wallet_id=wallet_id, public_key=public_key, plugins=plugins).serialize()

    @staticmethod
    def raw_create_transfer_msg(private_key: bytes, seqno: int, wallet_id: int, messages: typing.List[WalletMessage],
                                valid_until: typing.Optional[int] = None, op_code: int = 0) -> Cell:
        signing_message = Builder().store_uint(wallet_id, 32)
        if seqno == 0:
            signing_message.store_bits('1' * 32)  # bin(2**32 - 1)
        else:
            if valid_until is not None:
                signing_message.store_uint(valid_until, 32)
            else:
                signing_message.store_uint(int(time.time()) + 60, 32)
        signing_message.store_uint(seqno, 32)
        signing_message.store_uint(op_code, 8)
        for m in messages:
            signing_message.store_cell(m.serialize())
        signing_message = signing_message.end_cell()
        signature = sign_message(signing_message.hash, private_key)
        return Builder() \
            .store_bytes(signature) \
            .store_cell(signing_message) \
            .end_cell()

    @property
    def seqno(self) -> int:
        """
        :return: seqno taken from contract data
        """
        return WalletV4Data.deserialize(self.state.data.begin_parse()).seqno

    @property
    def wallet_id(self) -> int:
        """
        :return: wallet_id taken from contract data
        """
        return WalletV4Data.deserialize(self.state.data.begin_parse()).wallet_id

    @property
    def public_key(self) -> bytes:
        """
        :return: public_key taken from contract data
        """
        return WalletV4Data.deserialize(self.state.data.begin_parse()).public_key

    @property
    def plugins(self) -> Cell:
        return WalletV4Data.deserialize(self.state.data.begin_parse()).plugins

    async def get_plugin_list(self):
        """
        :return: plugins list from wallet's get method
        """
        return (await super().run_get_method(method='get_plugin_list', stack=[]))[0]

    async def is_plugin_installed(self, address: Address) -> bool:
        """
        :return: is plugin installed from wallet's get method
        """
        return bool((await super().run_get_method(method='is_plugin_installed', stack=[address.wc, address.hash_part]))[0])


class WalletV3R1(WalletV3):

    @classmethod
    async def from_data(cls, provider: LiteClient, public_key: bytes, wc: int = 0,
                        wallet_id: typing.Optional[int] = None, **kwargs):
        return await super().from_code_and_data(provider=provider, code=WALLET_V3_R1_CODE, public_key=public_key, wc=wc,
                                                wallet_id=wallet_id, **kwargs)

    @classmethod
    async def from_mnemonic(cls, provider: LiteClient, mnemonics: typing.Union[list, str], wc: int = 0,
                            wallet_id: typing.Optional[int] = None):
        if isinstance(mnemonics, str):
            mnemonics = mnemonics.split()
        assert mnemonic_is_valid(mnemonics), 'mnemonics are invalid!'
        _, private_key = mnemonic_to_private_key(mnemonics)
        return await super().from_private_key(provider, private_key, wc, wallet_id, 'v3r1')

    @classmethod
    async def create(cls, provider: LiteClient, wc: int = 0, wallet_id: typing.Optional[int] = None):
        """
        :param provider: provider
        :param wc: wallet workchain
        :param wallet_id: subwallet_id
        :return: mnemonics and Wallet instance of provided version
        """
        return super().create(provider=provider, wc=wc, wallet_id=wallet_id, version='v3r1')


class WalletV3R2(WalletV3):

    @classmethod
    async def from_data(cls, provider: LiteClient, public_key: bytes, wc: int = 0,
                        wallet_id: typing.Optional[int] = None, **kwargs):
        return await super().from_code_and_data(provider=provider, code=WALLET_V3_R2_CODE, public_key=public_key, wc=wc,
                                                wallet_id=wallet_id, **kwargs)

    @classmethod
    async def from_mnemonic(cls, provider: LiteClient, mnemonics: typing.Union[list, str], wc: int = 0,
                            wallet_id: typing.Optional[int] = None):
        if isinstance(mnemonics, str):
            mnemonics = mnemonics.split()
        assert mnemonic_is_valid(mnemonics), 'mnemonics are invalid!'
        _, private_key = mnemonic_to_private_key(mnemonics)
        return await super().from_private_key(provider, private_key, wc, wallet_id, 'v3r2')

    @classmethod
    async def create(cls, provider: LiteClient, wc: int = 0, wallet_id: typing.Optional[int] = None):
        """
        :param provider: provider
        :param wc: wallet workchain
        :param wallet_id: subwallet_id
        :return: mnemonics and Wallet instance of provided version
        """
        return super().create(provider=provider, wc=wc, wallet_id=wallet_id, version='v3r2')


class WalletV4R2(WalletV4):

    @classmethod
    async def from_data(cls, provider: LiteClient, public_key: bytes, wc: int = 0,
                        wallet_id: typing.Optional[int] = None, **kwargs):
        return await super().from_code_and_data(provider=provider, code=WALLET_V4_R2_CODE, public_key=public_key, wc=wc,
                                                wallet_id=wallet_id, **kwargs)

    @classmethod
    async def from_mnemonic(cls, provider: LiteClient, mnemonics: typing.Union[list, str], wc: int = 0,
                            wallet_id: typing.Optional[int] = None):
        if isinstance(mnemonics, str):
            mnemonics = mnemonics.split()
        assert mnemonic_is_valid(mnemonics), 'mnemonics are invalid!'
        _, private_key = mnemonic_to_private_key(mnemonics)
        return await super().from_private_key(provider, private_key, wc, wallet_id, 'v4r2')

    @classmethod
    async def create(cls, provider: LiteClient, wc: int = 0, wallet_id: typing.Optional[int] = None):
        """
        :param provider: provider
        :param wc: wallet workchain
        :param wallet_id: subwallet_id
        :return: mnemonics and Wallet instance of provided version
        """
        return super().create(provider=provider, wc=wc, wallet_id=wallet_id, version='v4r2')



import asyncio
import pytest

from pytoniq.adnl.adnl import AdnlTransport, Node


adnl = AdnlTransport(timeout=3)


@pytest.mark.asyncio
async def test_connection():

    # start adnl receiving server
    await adnl.start()

    # take peer from public config
    peer = Node('172.104.59.125', 14432, "/YDNd+IwRUgL0mq21oC0L3RxrS8gTu0nciSPUrhqR78=", adnl)
    await adnl.connect_to_peer(peer)

    # ask peer for something
    await peer.get_signed_address_list()

    # send pings to peer
    await asyncio.sleep(2)

    await peer.disconnect()

    # add another peer
    peer = Node('5.161.60.160', 12485, "jXiLaOQz1HPayilWgBWhV9xJhUIqfU95t+KFKQPIpXg=", adnl)

    # second way to connect to peer
    await peer.connect()

    # 2 adnl channels
    print(adnl.channels)

    # check for pings
    await asyncio.sleep(10)

    # stop adnl receiving server
    await adnl.close()



from ..contract import Contract, ContractError
from ...liteclient.client import LiteClient
from pytoniq_core.boc import Cell, Builder, HashMap
from pytoniq_core.boc.address import Address
from pytoniq_core.tlb.custom.nft import NftItemData

NFT_CODE = Cell.one_from_boc(b'\xb5\xee\x9crA\x02\r\x01\x00\x01\xd0\x00\x01\x14\xff\x00\xf4\xa4\x13\xf4\xbc\xf2\xc8\x0b\x01\x02\x01b\x02\x03\x02\x02\xce\x04\x05\x00\t\xa1\x1f\x9f\xe0\x05\x02\x01 \x06\x07\x02\x01 \x0b\x0c\x02\xd7\x0c\x88q\xc0$\x97\xc0\xf844\xc0\xc0\\l$\x97\xc0\xf8>\x90>\x90\x0c~\x80\x0c\\u\xc8~\x80\x0c~\x80\x0c<\x00\x81,\xe3\x85\x0c\x1b\x08\x8d\x14\x8c\xb1\xc1|\xb8e@~\x905\x0c\x04\x08\xfc\x00\xf8\x01\xb4\xc7\xf4\xcf\xe0\x84\x17\xf3\x0fE\x14\x8c.\xa3\xa1\xcc\x84\r\xd7\x8c\x90\x04\xf8\x0c\r\r\rM`\x84\x0b\xf2\xc9\xa8\x84\xae\xb8\xc0\x97\xc1!\x03\xfc\xbc \x08\t\x00\x11>\x91\x0c\x1c.\xbc\xb8S`\x01\xf6Q5\xc7\x05\xf2\xe1\x91\xfa@!\xf0\x01\xfa@\xd2\x001\xfa\x00\x82\n\xfa\xf0\x80\x1b\xa1!\x94S\x15\xa0\xa1\xde"\xd7\x0b\x01\xc3\x00 \x92\x06\xa1\x916\xe2 \xc2\xff\xf2\xe1\x92!\x8e>\x82\x10\x05\x13\x8d\x91\xc8P\t\xcf\x16P\x0b\xcf\x16q$I\x14TF\xa0p\x80\x10\xc8\xcb\x05P\x07\xcf\x16P\x05\xfa\x02\x15\xcbj\x12\xcb\x1f\xcb?"n\xb3\x94X\xcf\x17\x01\x912\xe2\x01\xc9\x01\xfb\x00\x10G\x94\x10*7[\xe2\n\x00rp\x82\x10\x8bw\x175\x05\xc8\xcb\xffP\x04\xcf\x16\x10$\x80@p\x80\x10\xc8\xcb\x05P\x07\xcf\x16P\x05\xfa\x02\x15\xcbj\x12\xcb\x1f\xcb?"n\xb3\x94X\xcf\x17\x01\x912\xe2\x01\xc9\x01\xfb\x00\x00\x82\x02\x8e5&\xf0\x01\x82\x10\xd52v\xdb\x107D\x00mqp\x80\x10\xc8\xcb\x05P\x07\xcf\x16P\x05\xfa\x02\x15\xcbj\x12\xcb\x1f\xcb?"n\xb3\x94X\xcf\x17\x01\x912\xe2\x01\xc9\x01\xfb\x00\x93024\xe2U\x02\xf0\x03\x00;;Q44\xcf\xfe\x90\x085\xd2p\x80&\x9f\xc0~\x905\x0c\x04\t\x04\x08\xf8\x0c\x1c\x16[[`\x00\x1d\x00\xf22\xcf\xd63\xc5\x80s\xc5\xb32{U \xbfu\x04\x1b')


class NftItem(Contract):

    @classmethod
    async def from_data(cls, provider: LiteClient, index: int, collection_address: Address, owner_address: Address, content: Cell, wc: int = 0, **kwargs) -> "NftItem":
        data = cls.create_data_cell(index, collection_address, owner_address, content)
        return await super().from_code_and_data(provider, wc, NFT_CODE, data, **kwargs)

    @staticmethod
    def create_data_cell(index: int, collection_address: Address, owner_address: Address, content: Cell) -> Cell:
        return NftItemData(index=index, collection_address=collection_address, owner_address=owner_address, content=content).serialize()

    @property
    def index(self) -> int:
        """
        :return: index taken from contract data
        """
        return NftItemData.deserialize(self.state.data.begin_parse()).index

    @property
    def collection_address(self) -> Address:
        """
        :return: collection_address taken from contract data
        """
        return NftItemData.deserialize(self.state.data.begin_parse()).collection_address

    @property
    def owner_address(self) -> Address:
        """
        :return: owner_address taken from contract data
        """
        return NftItemData.deserialize(self.state.data.begin_parse()).owner_address

    @property
    def content(self) -> Cell:
        """
        :return: old_queries taken from contract data
        """
        return NftItemData.deserialize(self.state.data.begin_parse()).content

    async def get_nft_data(self) -> bool:
        return await super().run_get_method(method='get_nft_data', stack=[])



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pytoniq",
    version="0.1.16",
    author="Maksim Kurbatov",
    author_email="cyrbatoff@gmail.com",
    description="TON Blockchain SDK",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages('.', exclude=['tests', 'examples']),
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries",
    ],
    url="https://github.com/yungwine/pytoniq",
    python_requires='>=3.9',
    py_modules=["pytoniq"],
    install_requires=[
        "pytoniq-core>=0.1.15",
        "requests>=2.31.0",
        "setuptools>=65.5.1",
    ]
)



import asyncio.exceptions
import base64
import time
import typing
import hashlib
import socket
import struct

import requests
from pytoniq_core.crypto.ciphers import Client, Server
from pytoniq_core.crypto.signature import verify_sign
from pytoniq_core.tl import TlGenerator

from .adnl import Node, AdnlTransport
from .overlay import OverlayNode, OverlayTransport


class DhtError(Exception):
    pass


class DhtValueNotFoundError(DhtError):
    pass


class DhtNode(Node):

    async def find_value(self, key: bytes, k: int = 6):
        data = {'key': key.hex(), 'k': k}
        return await self.transport.send_query_message('dht.findValue', data, self)

    async def store_value(self, value: dict):
        data = {'value': value}
        return await self.transport.send_query_message('dht.store', data, self)

    @classmethod
    def from_dict(cls, transport: AdnlTransport, data: dict, check_signature=True) -> "DhtNode":
        try:
            pub_k = bytes.fromhex(data['id']['key'])
            pub_k_b64 = base64.b64encode(pub_k)
        except ValueError:
            pub_k_b64 = data['id']['key']
            pub_k = base64.b64decode(pub_k_b64)
            data['id']['key'] = pub_k.hex()
        if isinstance(data['signature'], bytes):
            signature = data['signature']
        else:
            signature = base64.b64decode(data['signature'])
        data['signature'] = b''

        # check signature
        if check_signature:
            schemas = TlGenerator.with_default_schemas().generate()
            signed_message = schemas.serialize(schema=schemas.get_by_name('dht.node'), data=data)
            if not verify_sign(pub_k, signed_message, signature):
                raise Exception('invalid node signature!')

        node_addr = data['addr_list']['addrs'][0]
        host = socket.inet_ntoa(struct.pack('>i', node_addr['ip']))
        return cls(peer_host=host, peer_port=node_addr['port'], peer_pub_key=pub_k_b64, transport=transport)


class DhtClient:

    def __init__(self,
                 nodes: typing.List[DhtNode],
                 adnl_transport: AdnlTransport,
                 tl_schemas_path: typing.Optional[str] = None
                 ):
        self.adnl_transport: AdnlTransport = adnl_transport
        self.nodes_set: set = set(nodes)
        assert len(nodes) >= 1, 'expected at least 1 node in the list'
        if tl_schemas_path is None:
            self.schemas = TlGenerator.with_default_schemas().generate()
        else:
            self.schemas = TlGenerator(tl_schemas_path).generate()

    async def close(self):
        """
        disconnects to all known nodes
        :return:
        """
        for node in self.nodes_set:
            await node.disconnect()

    def get_dht_key_id_tl(self, id_: typing.Union[bytes, str], name: bytes = b'address', idx: int = 0):
        if isinstance(id_, str):
            id_ = bytes.fromhex(id_)
        dht_key_sch = self.schemas.get_by_name('dht.key')
        serialized = self.schemas.serialize(dht_key_sch, data={'id': id_.hex(), 'name': name, 'idx': idx})
        return hashlib.sha256(serialized).digest()

    @staticmethod
    def get_dht_key_id(id_: typing.Union[bytes, str], name: bytes = b'address', idx: int = 0):
        """
        Same as the method above but without using TlGenerator
        """
        if isinstance(id_, str):
            id_ = bytes.fromhex(id_)
        to_hash = b'\x8f\xdeg\xf6' + id_ + len(name).to_bytes(1, 'big') + name + idx.to_bytes(4, 'little')
        return hashlib.sha256(to_hash).digest()

    @staticmethod
    def find_distance_between_nodes(key_id_1: typing.Union[bytes, int], key_id_2: typing.Union[bytes, int]):
        if isinstance(key_id_1, bytes):
            key_id_1 = int.from_bytes(key_id_1, 'big')
        if isinstance(key_id_2, bytes):
            key_id_2 = int.from_bytes(key_id_2, 'big')
        return key_id_1 ^ key_id_2

    @classmethod
    def build_priority_list(cls, nodes: typing.Iterable, key_id: bytes) -> typing.List[DhtNode]:
        return sorted(nodes, key=lambda i: cls.find_distance_between_nodes(i.key_id, key_id))

    async def find_value(self, key: bytes, k: int = 6, timeout: int = 10):
        start_time = time.time()
        while True:
            if time.time() - start_time > timeout:  # TODO improve timeout
                raise asyncio.exceptions.TimeoutError()
            nodes = self.build_priority_list(self.nodes_set, key)

            for node in nodes:
                if not node.connected:
                    try:
                        await asyncio.wait_for(node.connect(), 1)
                    except asyncio.TimeoutError:
                        continue
                try:
                    resp = await node.find_value(key=key, k=k)
                except asyncio.exceptions.TimeoutError:
                    continue
                except Exception as e:
                    raise e  # ?

                resp = resp[0]
                if resp['@type'] == 'dht.valueNotFound':
                    new_nodes = resp['nodes']['nodes']
                    new_nodes_set = set()
                    for n in new_nodes:
                        new_nodes_set.add(DhtNode.from_dict(self.adnl_transport, n, True))
                    old_nodes = self.nodes_set.copy()
                    self.nodes_set = self.nodes_set.union(new_nodes_set)
                    if self.nodes_set == old_nodes:
                        raise DhtValueNotFoundError(f'value {key.hex()} not found')
                    break
                elif resp['@type'] == 'dht.valueFound':
                    return resp
                else:
                    raise DhtError(f'received unknown response type: {resp}')

    async def raw_store_value(self, dht_value: dict, try_find_after: bool = True):
        """
        dht.store value:dht.value = dht.Stored;

        :param dht_value: dict that represents `dht.value` TL scheme
        :param try_find_after: tries to find value in known peers after storage
        :return: bool was value stored or not
        """
        s = time.time()
        key = dht_value.get('key', {}).get('key', {})
        if key is None:
            raise DhtError(f'must provide dht.value dict to the method')
        key_id = self.get_dht_key_id(bytes.fromhex(key.get('id')), key.get('name'), key.get('idx'))
        nodes = self.build_priority_list(self.nodes_set, key_id)
        stored = False
        for node in nodes:
            if not node.connected:
                try:
                    await asyncio.wait_for(node.connect(), 2)
                except asyncio.TimeoutError:
                    continue
            try:
                resp = await node.store_value(dht_value)
                assert resp[0]['@type'] == 'dht.stored'
                stored = True
            except asyncio.exceptions.TimeoutError:
                continue
            except Exception as e:
                raise e  # ?
            stored = True

        if try_find_after:
            try:
                await self.find_value(key_id, timeout=10)
            except asyncio.exceptions.TimeoutError:
                return False
            except DhtValueNotFoundError:
                return False

        return stored

    @staticmethod
    def get_dht_key(id_: bytes, name: bytes = b'address', idx: int = 0):
        return {'id': id_.hex(), 'name': name, 'idx': idx}

    async def store_value(self, key: dict, value: bytes, private_key: bytes,
                          update_rule: typing.Literal['signature', 'anybody', 'overlayNodes'] = 'signature',
                          ttl: int = 30, try_find_after: bool = True):
        if update_rule != 'signature':
            raise DhtError('currently overlay is not supported')
        pk = Client(ed25519_private_key=private_key)
        key_description = {
            'key': key,
            'id': {
                '@type': 'pub.ed25519',
                'key': pk.ed25519_public.encode().hex()
            },
            'update_rule': self.schemas.get_by_name('dht.updateRule.' + update_rule).little_id(),
            'signature': b''
        }
        signature = pk.sign(self.schemas.serialize(self.schemas.get_by_name('dht.keyDescription'), key_description))

        data = {
            'key': key_description | {'signature': signature},
            'value': value,
            'ttl': int(time.time()) + ttl,
            'signature': b''
        }
        signature = pk.sign(self.schemas.serialize(self.schemas.get_by_name('dht.value'), data))

        data |= {'signature': signature}
        return await self.raw_store_value(data, try_find_after)

    async def get_overlay_nodes(self, overlay_id: typing.Union[bytes, str], overlay_transport: OverlayTransport):
        resp = await self.find_value(key=self.get_dht_key_id_tl(overlay_id, name=b'nodes'), timeout=30)
        nodes = resp['value']['value']['nodes']
        result = []
        for node in nodes:
            result.append(await self.get_overlay_node(node, overlay_transport))
        return result

    async def get_overlay_node(self, node: dict, overlay_transport: OverlayTransport) -> typing.Optional[OverlayNode]:
        """
        :param node: dict overlay.node TL schema
        :param overlay_transport:
        :return: OverlayNode or None
        """
        pub_k = bytes.fromhex(node['id']['key'])
        adnl_addr = Server('', 0, pub_key=pub_k).get_key_id()

        to_sign = self.schemas.serialize(
            schema=self.schemas.get_by_name('overlay.node.toSign'),
            data={'id': {'id': adnl_addr.hex()}, 'overlay': node['overlay'], 'version': node['version']}
        )

        if not verify_sign(pub_k, to_sign, node['signature']):
            raise Exception('invalid node signature!')

        try:
            resp = await self.find_value(key=self.get_dht_key_id_tl(id_=adnl_addr), timeout=5)
        except asyncio.TimeoutError:
            return None

        node_addr = resp['value']['value']['addrs'][0]
        host = socket.inet_ntoa(struct.pack('>i', node_addr['ip']))
        port = node_addr['port']
        pub_k = base64.b64encode(bytes.fromhex(resp['value']['key']['id']['key'])).decode()

        node = OverlayNode(peer_host=host, peer_port=port, peer_pub_key=pub_k, transport=overlay_transport)
        return node

    @classmethod
    def from_config(cls, config: dict, adnl_transport: AdnlTransport):
        nodes = []
        nodes_dict = config['dht']['static_nodes']['nodes']
        for node in nodes_dict:
            nodes.append(DhtNode.from_dict(adnl_transport, node))
        return cls(nodes, adnl_transport)

    @classmethod
    def from_mainnet_config(cls, adnl_transport: AdnlTransport):
        config = requests.get('https://ton.org/global-config.json').json()
        return cls.from_config(config, adnl_transport)

    @classmethod
    def from_testnet_config(cls, adnl_transport: AdnlTransport):
        config = requests.get('https://ton.org/testnet-global.config.json').json()
        return cls.from_config(config, adnl_transport)
